from . import merged_s2s
from .merged_s2s import *

from . import merged_s2_positions_cnf
from .merged_s2_positions_cnf import *

from . import merged_s2_positions_mlp
from .merged_s2_positions_mlp import *



import inspect
import textwrap

import pandas as pd
import strax
from strax.context import select_docs
import straxen


export, __all__ = strax.exporter()

ma_doc = (
    """
This is a straxen mini-analysis.
The method takes run_id as its only positional argument,
and additional arguments through keywords only.

The function requires the data types: {requires}.
Unless you specify this through data_kind = array keyword arguments,
this data will be loaded automatically.

The function takes the same selection arguments as context.get_array:
"""
    + select_docs
)

_hv_bokeh_initialized = False


@export
def mini_analysis(
    requires=tuple(), hv_bokeh=False, warn_beyond_sec=None, default_time_selection="touching"
):
    def decorator(f):
        parameters = inspect.signature(f).parameters

        def wrapped_f(context: strax.Context, run_id: str, **kwargs):
            # Validate arguments
            known_kwargs = (
                "time_range seconds_range time_within time_selection "
                "ignore_time_warning "
                "selection t_reference to_pe config"
            ).split()
            for k in kwargs:
                if k not in known_kwargs and k not in parameters:
                    # Python itself also raises TypeError for invalid kwargs
                    raise TypeError(f"Unknown argument {k} for {f.__name__}")

            if "config" in kwargs:
                context = context.new_context(config=kwargs["config"])
            if "config" in parameters:
                kwargs["config"] = context.config

            # Say magic words to enable holoviews
            if hv_bokeh:
                # Generally using globals is not great, but it would be
                # the same as doing a slow import on the top of this file
                # pylint: disable=global-statement
                global _hv_bokeh_initialized
                if not _hv_bokeh_initialized:
                    import holoviews

                    holoviews.extension("bokeh")
                    _hv_bokeh_initialized = True

            if "to_pe" in parameters and "to_pe" not in kwargs:
                to_pe = context.config["gain_model"]
                if isinstance(to_pe, str):
                    to_pe = straxen.URLConfig.evaluate_dry(to_pe, run_id=run_id)
                kwargs["to_pe"] = to_pe

            # Prepare selection arguments
            kwargs["time_range"] = context.to_absolute_time_range(
                run_id,
                targets=requires,
                **{k: kwargs.get(k) for k in ("time_range seconds_range time_within".split())},
            )
            kwargs.setdefault("time_selection", default_time_selection)
            kwargs.setdefault("selection", None)

            kwargs["t_reference"], _ = context.estimate_run_start_and_end(run_id, requires)

            if warn_beyond_sec is not None and not kwargs.get("ignore_time_warning"):
                tr = kwargs["time_range"]
                if tr is None:
                    sec_requested = float("inf")
                else:
                    sec_requested = (tr[1] - tr[0]) / int(1e9)
                if sec_requested > warn_beyond_sec:
                    tr_str = "the entire run" if tr is None else f"{sec_requested} seconds"
                    raise ValueError(
                        "The author of this mini analysis recommends "
                        f"not requesting more than {warn_beyond_sec} seconds. "
                        f"You are requesting {tr_str}. If you wish to proceed, "
                        "pass ignore_time_warning = True."
                    )

            # Load required data, if any
            if len(requires):
                deps_by_kind = strax.group_by_kind(requires, context=context)
                for dkind, dtypes in deps_by_kind.items():
                    if dkind in kwargs:
                        # Already have data, just apply cuts
                        kwargs[dkind] = strax.apply_selection(
                            kwargs[dkind],
                            selection=kwargs["selection"],
                            time_range=kwargs["time_range"],
                            time_selection=kwargs["time_selection"],
                        )
                    else:
                        kwargs[dkind] = context.get_array(
                            run_id,
                            dtypes,
                            selection=kwargs["selection"],
                            time_range=kwargs["time_range"],
                            time_selection=kwargs["time_selection"],
                            # Arguments for new context, if needed
                            config=kwargs.get("config"),
                            register=kwargs.get("register"),
                            storage=kwargs.get("storage", tuple()),
                            progress_bar=False,
                        )

                # If user did not give time kwargs, but the function expects
                # a time_range, try to add one based on the time range of the data
                base_dkind = list(deps_by_kind.keys())[0]
                x = kwargs[base_dkind]
                if len(x) and kwargs.get("time_range") is None:
                    x0 = x.iloc[0] if isinstance(x, pd.DataFrame) else x[0]
                    try:
                        kwargs.setdefault("time_range", (x0["time"], strax.endtime(x).max()))

                    except AttributeError:
                        # If x is a holoviews dataset, this will fail.
                        pass

            if "seconds_range" in parameters:
                if kwargs.get("time_range") is None:
                    scr = None
                else:
                    scr = tuple(
                        [(t - kwargs["t_reference"]) / int(1e9) for t in kwargs["time_range"]]
                    )
                kwargs.setdefault("seconds_range", scr)

            kwargs.setdefault("run_id", run_id)
            kwargs.setdefault("context", context)

            if "kwargs" in parameters:
                # Likely this will be passed to another mini-analysis
                to_pass = kwargs
                # Do not pass time_range and seconds_range both (unless explicitly requested)
                # strax does not like that
                if "seconds_range" in to_pass and "seconds_range" not in parameters:
                    del to_pass["seconds_range"]
                if "time_within" in to_pass and "time_within" not in parameters:
                    del to_pass["time_within"]
            else:
                # Pass only arguments the function wants
                to_pass = {k: v for k, v in kwargs.items() if k in parameters}
            return f(**to_pass)

        wrapped_f.__name__ = f.__name__

        if hasattr(f, "__doc__") and f.__doc__:
            doc_lines = f.__doc__.splitlines()
            wrapped_f.__doc__ = doc_lines[0] + "\n" + textwrap.dedent("\n".join(doc_lines[1:]))
        else:
            wrapped_f.__doc__ = (
                "Straxen mini-analysis for which someone was too lazy to write a proper docstring"
            )

        wrapped_f.__doc__ += ma_doc.format(requires=", ".join(requires))

        strax.Context.add_method(wrapped_f)
        return wrapped_f

    return decorator



import numpy as np
import strax


class EventAmbience(strax.Plugin):
    """Save Ambience of the main S1 and main S2 in the event.

    References:
        * v0.0.4 reference: xenon:xenonnt:ac:prediction:shadow_ambience
        * v0.1.0 reference: xenon:xenonnt:analysis:redefine_n_competing

    """

    __version__ = "0.1.0"
    depends_on = ("event_basics", "peak_ambience")
    provides = "event_ambience"

    @property
    def origin_dtype(self):
        return ["lh_before", "s0_before", "s1_before", "s2_before", "s2_near"]

    def infer_dtype(self):
        dtype = []
        for ambience in self.origin_dtype:
            dtype.append(
                (
                    (f"Number of {' '.join(ambience.split('_'))} main S1", f"s1_n_{ambience}"),
                    np.int16,
                )
            )
            dtype.append(
                (
                    (f"Number of {' '.join(ambience.split('_'))} main S2", f"s2_n_{ambience}"),
                    np.int16,
                )
            )
        dtype += [
            (
                ("Sum of small hits and peaks before main S1", "s1_s_before"),
                np.float32,
            ),
            (
                ("Sum of small hits and peaks before main S2", "s2_s_before"),
                np.float32,
            ),
        ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)

        # 1. Initialization, ambience is set to be the lowest possible value
        result = np.zeros(len(events), self.dtype)
        # 2. Assign peaks features to main S1, main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for idx, main_peak in zip([event["s1_index"], event["s2_index"]], ["s1_", "s2_"]):
                if idx >= 0:
                    for ambience in self.origin_dtype:
                        result[f"{main_peak}n_{ambience}"][event_i] = sp[f"n_{ambience}"][idx]
                    result[f"{main_peak}s_before"][event_i] = sp["s_before"][idx]

        # 3. Set time and endtime for events
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)
        return result



import strax
import numpy as np
from itertools import chain

export, __all__ = strax.exporter()


@export
class nVeto_reflectivity(strax.Plugin):
    """Plugin which computes parameters used for the Reflectivity Monitor for the Neutron Veto.

    From the raw records, the start time of each trigger window is extracted (by taking the smallest
    "time" of all the channels)

    Using the hitlets, the delta times wrt the beginning of the trigger are computed

    Cuts can be made in area (of all the hitlets), time (since the beginning of the trigger window)
    and number of contributing PMTs (in each window)

    The returned values are the delta times (used for the reflectivity), PMT channels and areas

    """

    __version__ = "0.1.0"

    parallel = "process"
    rechunk_on_save = True
    compressor = "zstd"

    depends_on = ("raw_records_nv", "hitlets_nv")

    provides = "ref_mon_nv"
    data_kind = "ref_mon_nv"

    dtype = strax.time_fields + [
        (("Time with respect to the precedent trigger [ns]", "time_since_trigger"), np.int32),
        (("Channel of the hit", "channel"), np.int16),
        (("Area of the hit [PE]", "area"), np.float32),
        (("Number of event", "number"), np.int32),
    ]

    def compute(self, raw_records_nv, hitlets_nv):
        split_time = 5000

        # find the trigger times
        time_unique = np.unique(raw_records_nv["time"])
        time_diff = np.diff(time_unique, prepend=time_unique[0])
        time_unique_split = np.split(time_unique, np.where(time_diff >= split_time)[0])
        trigger_times = np.array([i[0] for i in time_unique_split])

        # split the hits in trigger windows
        difference = np.diff(hitlets_nv["time"], prepend=hitlets_nv["time"][0])
        hits_split = np.split(hitlets_nv, np.where(difference > split_time)[0])
        numbers = [np.repeat(i, len(hits_split[i])) for i in range(len(hits_split))]

        # compute the time differences
        times_wrt_trigger = [
            hits_split[i]["time"]
            - trigger_times[np.argmin(np.abs(trigger_times - hits_split[i][0]["time"]))]
            for i in range(len(hits_split))
        ]

        # reshape the arrays
        hits_split = np.array(list(chain.from_iterable(hits_split)))
        times_wrt_trigger = np.array(list(chain.from_iterable(times_wrt_trigger)))
        numbers = np.array(list(chain.from_iterable(numbers)))

        # fill the delta_times
        delta_times = np.zeros(len(hits_split), dtype=self.dtype)
        delta_times["time"] = hits_split["time"]
        delta_times["endtime"] = hits_split["time"] + hits_split["length"] * hits_split["dt"]
        delta_times["time_since_trigger"] = times_wrt_trigger
        delta_times["area"] = hits_split["area"]
        delta_times["channel"] = hits_split["channel"]
        delta_times["number"] = numbers

        return delta_times



import strax
import straxen
from ._peaklet_positions_base import PeakletPositionsBase


export, __all__ = strax.exporter()


@export
class PeakletPositionsMLP(PeakletPositionsBase):
    """Multilayer Perceptron (MLP) neural net for position reconstruction."""

    __version__ = "0.0.0"
    child_plugin = True
    algorithm = "mlp"
    provides = "peaklet_positions_mlp"
    gc_collect_after_compute = True

    tf_model_mlp = straxen.URLConfig(
        default="tf://resource://xnt_mlp_SR0_mix_2000031_2000021_20211211.keras?fmt=abs_path",
        help=(
            'MLP model. Should be opened using the "tf" descriptor. '
            'Set to "None" to skip computation'
        ),
        cache=3,
    )



import strax
import numpy as np
from straxen.plugins.online_monitor_nv.online_monitor_nv import OnlineMonitorNV, veto_monitor_dtype

export, __all__ = strax.exporter()


@export
class OnlineMonitorMV(OnlineMonitorNV):
    __doc__ = (OnlineMonitorNV.__doc__ or "").replace("_nv", "_mv").replace("nVeto", "muVeto")
    depends_on = ("hitlets_mv", "events_mv")
    provides = "online_monitor_mv"
    data_kind = "online_monitor_mv"

    # Needed in case we make again an muVETO child.
    ends_with = "_mv"
    child_plugin = True

    __version__ = "0.0.2"

    def infer_dtype(self):
        self.channel_range = self.channel_map["mv"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_monitor_dtype(self.ends_with, self.n_channel, self.events_area_nbins)

    def compute(self, hitlets_mv, events_mv, start, end):
        events_mv = np.copy(events_mv)
        return super().compute(hitlets_mv, events_mv, start, end)



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class DistinctChannels(strax.LoopPlugin):
    """Compute the number of contributing PMTs that contribute to the alt_s1 but not to the main
    S1."""

    __version__ = "0.1.1"
    depends_on = ("event_basics", "peaks")
    loop_over = "events"
    dtype = [
        (
            "alt_s1_distinct_channels",
            np.int32,
            "Number of PMTs contributing to the secondary S1 that do not contribute to the main S1",
        ),
    ] + strax.time_fields

    def compute_loop(self, event, peaks):
        if event["alt_s1_index"] == -1:
            n_distinct = 0
        else:
            s1_a = peaks[event["s1_index"]]
            s1_b = peaks[event["alt_s1_index"]]
            s1_a_peaks = np.nonzero((s1_a["area_per_channel"] > 0) * 1)
            s1_b_peaks = np.nonzero((s1_b["area_per_channel"] > 0) * 1)
            n_distinct = 0
            for channel in range(len(s1_b_peaks[0])):
                if s1_b_peaks[0][channel] not in s1_a_peaks[0]:
                    n_distinct += 1

        return dict(
            alt_s1_distinct_channels=n_distinct, time=event["time"], endtime=event["endtime"]
        )



import numpy as np
import pandas as pd
import datetime
import utilix
import strax
import straxen

from straxen.plugins.aqmon_hits.aqmon_hits import AqmonChannels
from scipy.interpolate import interp1d


class GpsSync(strax.OverlapWindowPlugin):
    """Correct the event times to GPS time.

    1. Finds the TTL GPS pulses coming into the AM from the gps
    module and their pairs coming from the module for the
    correspondent run.
    2. Corrects the timestamp of all events by linearly interpolating
    between the previous and next sync pulses.

    """

    __version__ = "0.2.1"
    depends_on = ("aqmon_hits", "event_basics")
    provides = "gps_sync"
    data_kind = "events"

    gps_channel = AqmonChannels.GPS_SYNC_AM
    _is_tpc = True  # If true tpc_internal_delay is subtracted from final
    # GPS results. Has to be set to False when inherited
    # for the neutron-veto.

    tpc_internal_delay = straxen.URLConfig(
        default={"0": 4917, "020380": 10137, "043040": 0},
        type=dict,
        track=True,
        help="Internal delay between aqmon and regular TPC channels ins [ns]",
    )

    gps_truth_vicinity = straxen.URLConfig(
        default=4.5,
        type=float,
        track=True,
        help=(
            "+/- time interval in [s] in which the GPS truth should be searched for each GPS pulse."
        ),
    )

    dtype = strax.time_fields + [(("GPS absolute time [ns]", "t_gps"), np.int64)]

    def get_window_size(self):
        # Use a large window to ensure that multiple GPS pulses are found
        # within the current chunk. Specified window is in nanoseconds.
        return int(120 * straxen.units.s)

    def setup(self):
        # Load GPS-module pulses
        self.gps_times = self.load_gps_array()
        self.delay = 0
        if self._is_tpc:
            self.delay = self.get_delay()

    def load_gps_array(self):
        """Function which load GPS time information from data base and converts timestamps back into
        nanoseconds unix time."""
        gps_info = self.gps_times_from_runid(self.run_id)
        gps_info["pulse_time"] = np.int64(gps_info["gps_sec"] * 1e9) + np.int64(gps_info["gps_ns"])
        gps_array = strax.stable_sort(gps_info["pulse_time"])
        return gps_array

    def gps_times_from_runid(self, run_id):
        """Fetches the mongodb looking for the gps_sync collection for timestamps between the start
        and end times of a given run_id."""
        rundb = utilix.xent_collection()
        gps_times = utilix.xent_collection(collection="gps_sync", database="xenonnt")

        if isinstance(run_id, str):
            run_id = int(run_id)

        query = {"number": run_id}

        doc = rundb.find_one(query, projection={"start": 1, "end": 1})

        assert not (doc is None), "No match for run_id %s when computing GPS times." % self.run_id

        start_t = doc["start"].replace(tzinfo=datetime.timezone.utc).timestamp()
        end_t = doc["end"].replace(tzinfo=datetime.timezone.utc).timestamp()

        query = {"gps_sec": {"$gte": start_t - 11, "$lte": end_t + 11}, "channel": 0}

        return pd.DataFrame(gps_times.find(query))

    @staticmethod
    def match_gps_pulses_with_truth(gps_pulse_time, gps_truth_time, vicinity):
        """Function which matches GPS truth from the file stored in the rundb with the corresponding
        GPS pulse recorded by the acquisition monitor.

        :param gps_pulse_time: Unix ns timestamps corresponding to GPS pulse starts
        :param gps_truth_time: Unix ns timestamps of the GPS pulses
        :param vicinity: +/- vicinity in [s] for which GPS pulse is matched with GPS truth.

        """
        _gps_pulses = np.zeros(len(gps_pulse_time), strax.time_fields)
        _gps_pulses["time"] = gps_pulse_time - int(vicinity * straxen.units.s)
        _gps_pulses["endtime"] = gps_pulse_time + int(vicinity * straxen.units.s)

        _gps_truth = np.zeros(len(gps_truth_time), strax.time_fields)
        _gps_truth["time"] = gps_truth_time
        _gps_truth["endtime"] = gps_truth_time + 1

        truth_in_pulse_index = strax.fully_contained_in(_gps_truth, _gps_pulses)

        matched_gps_truth = gps_truth_time[truth_in_pulse_index >= 0]
        truth_in_pulse_index = truth_in_pulse_index[truth_in_pulse_index >= 0]

        matched_gps_pulses = gps_pulse_time[truth_in_pulse_index]

        return matched_gps_truth, matched_gps_pulses

    def compute_time_array(self, l_daq_sync, l_gps_sync, l_daq_evt):
        """Function which computes for each event the corresponding GPS time.

        :param l_daq_sync: Unix time ns timestamps for the recorded gps pulses
        :param l_gps_sync: Unix time ns timestamps for the GPS truth.
        :param l_daq_evt: Unix time ns timestamps for the events.

        """
        _has_no_gps_but_events = (len(l_daq_sync) == 0 or len(l_gps_sync) == 0) and len(l_daq_evt)
        if _has_no_gps_but_events:
            raise ValueError(
                f'Cannot compute GPS correction for "{self.run_id}". '
                f"There were {len(l_gps_sync)} GPS found in the run-docs, "
                "which are marching recorded GPS pulses, "
                f"with {len(l_daq_evt)} events."
            )

        _has_no_events = not len(l_daq_evt)
        if _has_no_events:
            return np.zeros(0, dtype=np.int64)

        # Important before interpolate convert times in unix ns times
        # into ns since run start, otherwise interpolation gets bit
        # issue with the conversion from int64 to float64.
        first_gps_time = l_daq_sync[0]

        interp_gps = interp1d(
            l_daq_sync - first_gps_time,
            l_gps_sync - first_gps_time,
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
        event_times_gps = interp_gps(l_daq_evt - first_gps_time)

        # Now convert back into int64 and add offset again
        event_times_gps = event_times_gps.astype(np.int64)
        return event_times_gps + first_gps_time

    def load_aqmon_array(self, hits):
        gps_hits = hits[hits["channel"] == self.gps_channel]
        aqmon_array = gps_hits["time"]
        return aqmon_array

    def get_delay(self):
        delay = 0
        for run_id, _delay in self.tpc_internal_delay.items():
            if int(self.run_id) >= int(run_id):
                delay = _delay
        return delay

    def compute(self, aqmon_hits, events):
        hits = aqmon_hits
        evts = events

        # Load pulses from aqmon
        gps_pulse_times = self.load_aqmon_array(hits)

        # Match GPS pulses with GPS truth, interpolate and extrapolate results
        # in order to include data at run boundaries.
        # Since the clock drift is tiny an extrapolation should be sufficient.
        matched_gps_truth, matched_gps_pulses = self.match_gps_pulses_with_truth(
            gps_pulse_times, self.gps_times, self.gps_truth_vicinity
        )

        t_events_gps = self.compute_time_array(matched_gps_pulses, matched_gps_truth, evts["time"])

        ans = dict()
        ans["time"] = evts["time"]
        ans["endtime"] = evts["endtime"]
        ans["t_gps"] = t_events_gps - self.delay

        return ans


class GpsSyncDAQVeto(GpsSync):
    """Plugin which computes veto_intervals using GPS times.

    Required to compute correctly total lifetime loss of experiment.

    """

    __version__ = "0.0.1"
    depends_on = ("aqmon_hits", "veto_intervals")
    provides = "veto_intervals_gps_sync"
    data_kind = "veto_intervals"
    child_plugin = True

    def compute(self, aqmon_hits, veto_intervals):
        res = super().compute(aqmon_hits, veto_intervals)
        return res


class GpsSync_nv(GpsSync):
    """Computes absolute GPS time for nveto data."""

    adc_threshold_nim_signal = straxen.URLConfig(
        default=500, type=int, track=True, help="Threshold in [adc] to search for the NIM signal"
    )
    # This value is only valid for SR0:
    epsilon_offset_nv_gps_pulse = straxen.URLConfig(
        default=76, type=int, track=True, help="Measured missing offset for nveto in [ns]"
    )

    __version__ = "0.1.0"
    depends_on = ("raw_records_aqmon_nv", "events_nv")
    provides = "events_gps_nv"
    data_kind = "events_nv"

    gps_channel = AqmonChannels.GPS_SYNC_NV
    _is_tpc = False

    dtype = strax.time_fields + [
        (("GPS absolute time [ns]", "time_sync_gps"), np.int64),
    ]

    def setup(self):
        super().setup()

    @staticmethod
    def get_nim_edge(raw_records, threshold=500):
        records = strax.raw_to_records(raw_records)
        strax.baseline(records)
        hits = strax.find_hits(records, min_amplitude=threshold)
        return hits

    def compute(self, raw_records_aqmon_nv, events_nv):
        rr_nv = raw_records_aqmon_nv
        _mask_nveto = rr_nv["channel"] == self.gps_channel
        hits_nv = self.get_nim_edge(rr_nv[_mask_nveto], self.adc_threshold_nim_signal)
        nveto_extra_offset = 0
        if not len(hits_nv):
            # During SR0 sync signal was not recorded properly for the
            # neutron-veto, hence take waveform itself as "hits".
            _mask_nveto &= rr_nv["record_i"] == 0
            nveto_extra_offset = self.epsilon_offset_nv_gps_pulse
            hits_nv = rr_nv[_mask_nveto]
        hits_nv["time"] += nveto_extra_offset

        res = super().compute(aqmon_hits=hits_nv, events=events_nv)
        gps_nv = np.zeros(len(res["time"]), self.dtype)
        gps_nv["time"] = res["time"]
        gps_nv["endtime"] = res["endtime"]
        gps_nv["time_sync_gps"] = res["t_gps"]
        return gps_nv


class GpsSync_mv(GpsSync):
    """Computes absolute GPS time for mveto data."""

    adc_threshold_nim_signal = straxen.URLConfig(
        default=500, type=int, track=True, help="Threshold in [adc] to search for the NIM signal"
    )

    __version__ = "0.1.0"
    depends_on = ("raw_records_aux_mv", "events_mv")
    provides = "events_gps_mv"
    data_kind = "events_mv"

    gps_channel = AqmonChannels.GPS_SYNC_MV
    _is_tpc = False

    dtype = strax.time_fields + [
        (("GPS absolute time [ns]", "time_sync_gps"), np.int64),
    ]

    def setup(self):
        super().setup()

    @staticmethod
    def get_nim_edge(raw_records, threshold=500):
        records = strax.raw_to_records(raw_records)
        strax.baseline(records)
        hits = strax.find_hits(records, min_amplitude=threshold)
        return hits

    def compute(self, raw_records_aux_mv, events_mv):
        rr_mv = raw_records_aux_mv
        _mask_nveto = rr_mv["channel"] == self.gps_channel
        hits_mv = self.get_nim_edge(rr_mv[_mask_nveto], self.adc_threshold_nim_signal)

        res = super().compute(aqmon_hits=hits_mv, events=events_mv)
        gps_mv = np.zeros(len(res["time"]), self.dtype)
        gps_mv["time"] = res["time"]
        gps_mv["endtime"] = res["endtime"]
        gps_mv["time_sync_gps"] = res["t_gps"]
        return gps_mv



import strax
import straxen
import numpy as np
from immutabledict import immutabledict

export, __all__ = strax.exporter()


@export
class nVETOEventWaveform(strax.Plugin):
    """Plugin which computes the summed waveform as well as some shape properties of the NV
    events."""

    __version__ = "0.0.2"

    depends_on = "events_nv", "records_nv"
    provides = "event_waveform_nv"
    data_kind = "events_nv"
    compressor = "zstd"

    gain_model_nv = straxen.URLConfig(
        default=(
            "list-to-array://"
            "xedocs://pmt_area_to_pes"
            "?as_list=True"
            "&sort=pmt"
            "&detector=neutron_veto"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
            "&attr=value"
        ),
        infer_type=False,
        help="PMT gain model. Specify as (model_type, model_config, nT = True)",
    )

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) channel number",
    )

    def infer_dtype(self):
        return veto_event_waveform_dtype()

    def setup(self):
        self.channel_range = self.channel_map["nveto"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1

        to_pe = self.gain_model_nv

        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0] :] = to_pe[:]

    def compute(self, events_nv, records_nv, start, end):
        events_waveform = np.zeros(len(events_nv), self.dtype)

        # Compute shape like properties:
        _tmp_events = np.zeros(len(events_nv), dtype=_temp_event_data_type())
        strax.copy_to_buffer(events_nv, _tmp_events, "_temp_nv_evts_wf_cpy")
        _tmp_events["length"] = (events_nv["endtime"] - events_nv["time"]) // 2
        _tmp_events["dt"] = 2
        strax.simple_summed_waveform(records_nv, _tmp_events, self.to_pe)
        strax.compute_properties(_tmp_events)

        strax.copy_to_buffer(_tmp_events, events_waveform, "_temp_nv_evts_cpy")
        events_waveform["range_50p_area"] = _tmp_events["width"][:, 5]
        events_waveform["range_90p_area"] = _tmp_events["width"][:, 9]
        events_waveform["rise_time"] = -_tmp_events["area_decile_from_midpoint"][:, 1]
        del _tmp_events

        return events_waveform


def veto_event_waveform_dtype(
    n_samples_wf: int = 200,
) -> list:
    dtype = strax.time_dt_fields + [
        (("Waveform data in PE/sample (not PE/ns!)", "data"), np.float32, n_samples_wf),
        (("Width (in ns) of the central 50% area of the peak", "range_50p_area"), np.float32),
        (("Width (in ns) of the central 90% area of the peak", "range_90p_area"), np.float32),
        (("Time between 10% and 50% area quantiles [ns]", "rise_time"), np.float32),
    ]
    return dtype


def _temp_event_data_type(n_samples_wf: int = 150, n_widths: int = 11) -> list:
    """Temp.

    data type which adds field required to use some of the functions used to compute the shape of
    the summed waveform for the TPC.

    """
    dtype = veto_event_waveform_dtype()
    dtype += [
        (
            ("Dummy, total area of all hitlets in event [pe]", "area"),
            np.float32,
        ),
        (
            ("Dummy top waveform data in PE/sample (not PE/ns!)", "data_top"),
            np.float32,
            n_samples_wf,
        ),
        (
            ("Dummy first waveform data in PE/sample (not PE/ns!)", "data_start"),
            np.float32,
            n_samples_wf,
        ),
        (("Weighted average center time of the peak [ns]", "center_time"), np.int64),
        (("Weighted relative median time of the peak [ns]", "median_time"), np.float32),
        (("Peak widths in range of central area fraction [ns]", "width"), np.float32, n_widths),
        (
            ("Peak widths: time between nth and 5th area decile [ns]", "area_decile_from_midpoint"),
            np.float32,
            n_widths,
        ),
    ]

    return dtype



import io
import os
from warnings import warn
from typing import Tuple
import tarfile
from immutabledict import immutabledict
from importlib import import_module

import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


nt_test_run_id = "012882"


@export
def download_test_data(
    test_data="https://raw.githubusercontent.com/XENONnT/strax_auxiliary_files/353b2c60a01e96f67e4ba544ce284bd91241964d/strax_files/strax_test_data_straxv1.1.0.tar",  # noqa
):
    """Downloads strax test data to strax_test_data in the current directory."""
    blob = straxen.common.get_resource(test_data, fmt="binary")
    f = io.BytesIO(blob)
    tf = tarfile.open(fileobj=f)
    tf.extractall()


@export
def _overwrite_testing_function_file(function_file):
    """For testing purposes allow this function file to be loaded from HOME/testing_folder."""
    if not _is_on_pytest():
        # If we are not on a pytest, never try using a local file.
        return function_file

    home = os.environ.get("HOME")
    if home is None:
        # Impossible to load from non-existent folder
        return function_file

    testing_file = os.path.join(home, function_file)

    if os.path.exists(testing_file):
        # For testing purposes allow loading from 'home/testing_folder'
        warn(
            f"Using local function: {function_file} from {testing_file}! "
            "If you are not integrated testing on github you should "
            "absolutely remove this file. (See #559)"
        )
        function_file = testing_file

    return function_file


def is_installed(module):
    try:
        import_module(module)
        return True
    except ModuleNotFoundError:
        return False


@export
def _is_on_pytest():
    """Check if we are on a pytest."""
    return "PYTEST_CURRENT_TEST" in os.environ


@export
def mongo_uri_not_set():
    return "TEST_MONGO_URI" not in os.environ


def _get_fake_daq_reader():
    class DAQReader(straxen.DAQReader):
        """Dummy version of the DAQ reader to make sure that all the testing data produced here will
        have a different lineage."""

        __version__ = "MOCKTESTDATA"

    return DAQReader


def nt_test_context(
    target_context="xenonnt_online", deregister=(), keep_default_storage=False, **kwargs
) -> strax.Context:
    """Get a dummy context with full nt-like data simulated data (except aqmon) to allow testing
    plugins.

    :param target_context: Which contexts from straxen.contexts to test
    :param deregister: a list of plugins from the context
    :param keep_default_storage: if to True, keep the default context storage. Usually, you don't
        need this since all the data will be stored in a separate test data folder.
    :param kwargs: Any kwargs are passed to the target-context
    :return: a context

    """
    if not straxen.utilix_is_configured(warning_message=False):
        kwargs.setdefault("_database_init", False)

    st = getattr(straxen.contexts, target_context)(**kwargs)
    st.set_config(
        {"diagnose_sorting": True, "diagnose_overlapping": True, "store_per_channel": True}
    )
    st.register(_get_fake_daq_reader())
    download_test_data(
        "https://raw.githubusercontent.com/XENONnT/"
        "strax_auxiliary_files/"
        "f0d177401e11408b273564f0e29df77528e83d26/"
        "strax_files/"
        "012882-raw_records-z7q2d2ye2t.tar"
    )
    if keep_default_storage:
        st.storage += [
            strax.DataDirectory("./strax_test_data", deep_scan=True, provide_run_metadata=True)
        ]
    else:
        st.storage = [
            strax.DataDirectory("./strax_test_data", deep_scan=True, provide_run_metadata=True)
        ]
    assert st.is_stored(nt_test_run_id, "raw_records"), os.listdir(st.storage[-1].path)

    to_remove = list(deregister)
    for plugin in to_remove:
        del st._plugin_class_registry[plugin]

    # Change the led_plugin default_run_comments to be compatible
    # with the test run_id in straxen.
    st.set_config({"default_run_comments": ["S1-only"]})

    return st


def create_unique_intervals(size, time_range=(0, 40), allow_zero_length=True):
    """Hypothesis stragtegy which creates unqiue time intervals.

    :param size: Number of intervals desired. Can be less if non-unique intervals are found.
    :param time_range: Time range in which intervals should be.
    :param allow_zero_length: If true allow zero length intervals.

    """
    from hypothesis import strategies

    strat = strategies.lists(
        elements=strategies.integers(*time_range), min_size=size * 2, max_size=size * 2
    ).map(lambda x: _convert_to_interval(x, allow_zero_length))
    return strat


def _convert_to_interval(time_stamps, allow_zero_length):
    time_stamps = strax.stable_sort(time_stamps)
    intervals = np.zeros(len(time_stamps) // 2, strax.time_dt_fields)
    intervals["dt"] = 1
    intervals["time"] = time_stamps[::2]
    intervals["length"] = time_stamps[1::2] - time_stamps[::2]

    if not allow_zero_length:
        intervals = intervals[intervals["length"] > 0]
    return np.unique(intervals)


@strax.takes_config(
    strax.Option("secret_time_offset", default=0, track=False),
    strax.Option("recs_per_chunk", default=10, track=False),
    strax.Option(
        "n_chunks",
        default=2,
        track=False,
        help="Number of chunks for the dummy raw records we are writing here",
    ),
    strax.Option(
        "channel_map",
        track=False,
        type=immutabledict,
        help="frozendict mapping subdetector to (min, max) channel number.",
    ),
)
class DummyRawRecords(strax.Plugin):
    """Provide dummy raw records for the mayor raw_record types."""

    provides = straxen.daqreader.DAQReader.provides
    parallel = "process"
    depends_on: Tuple = tuple()
    data_kind = immutabledict(zip(provides, provides))
    rechunk_on_save = False
    dtype = {p: strax.raw_record_dtype() for p in provides}

    def setup(self):
        self.channel_map_keys = {
            "he": "he",
            "nv": "nveto",
            "aqmon": "aqmon",
            "aux_mv": "aux_mv",
            "s_mv": "mv",
        }  # s_mv otherwise same as aux in endswith

    def source_finished(self):
        return True

    def is_ready(self, chunk_i):
        return chunk_i < self.config["n_chunks"]

    def compute(self, chunk_i):
        t0 = chunk_i + self.config["secret_time_offset"]
        if chunk_i < self.config["n_chunks"] - 1:
            # One filled chunk
            r = np.zeros(self.config["recs_per_chunk"], self.dtype["raw_records"])
            r["time"] = t0
            r["length"] = r["dt"] = 1
            r["channel"] = np.arange(len(r))
        else:
            # One empty chunk
            r = np.zeros(0, self.dtype["raw_records"])

        res = {}
        for p in self.provides:
            rr = np.copy(r)
            # Add detector specific channel offset:
            for key, channel_key in self.channel_map_keys.items():
                if p.endswith(key):
                    first_channel, last_channel = self.config["channel_map"][channel_key]
                    rr["channel"] += first_channel
                    if key == "aqmon":
                        # explicitly clip these channels
                        # as we have an additional check higher in the chain
                        first_channel = int(min(straxen.AqmonChannels))
                        last_channel = int(max(straxen.AqmonChannels))

                    rr = rr[(rr["channel"] >= first_channel) & (rr["channel"] < last_channel)]
            res[p] = self.chunk(start=t0, end=t0 + 1, data=rr, data_type=p)
        return res



import strax
import straxen


def check_global_version_wrapper(func):
    """Check whether all configs for superruns-allowed plugins are the same for subruns."""

    def wrapper(self, run_id, targets=tuple(), **kwargs):
        if (
            not self.context_config["check_global_version_configs"]
            or "xedocs_version" not in self.context_config
            or self.context_config["xedocs_version"] is None
            or self.context_config["xedocs_version"] == "global_ONLINE"
        ):
            return func(self, run_id=run_id, targets=targets, **kwargs)

        configs = self.versioned_configs(run_id, targets)
        if not configs:
            return func(self, run_id=run_id, targets=targets, **kwargs)

        for key, value in configs.items():
            url, plugin = value
            arg, extra_kwargs = straxen.URLConfig.split_url_kwargs(url)
            if extra_kwargs["version"] == "ONLINE":
                raise ValueError(
                    f"The global version is set to be {self.context_config['xedocs_version']}. "
                    f"But {plugin.__class__.__name__} is still using ONLINE version "
                    f"config {key}, which is {url}."
                )

        return func(self, run_id=run_id, targets=targets, **kwargs)

    return wrapper


# Manually decorate context class
_strax_context_check_global_version_decorated = True
if "check_global_version_configs" not in strax.Context.takes_config:
    strax.Context = strax.takes_config(
        strax.Option(
            name="check_global_version_configs",
            default=True,
            type=bool,
            help=(
                "If True, check whether version=ONLINE is not used "
                "when global version is not global_ONLINE."
            ),
        ),
    )(strax.Context)
    _strax_context_check_global_version_decorated = False
if not _strax_context_check_global_version_decorated:
    # Overwrite get_components method
    strax.Context.get_components = check_global_version_wrapper(strax.Context.get_components)


@strax.Context.add_method
def versioned_configs(self, run_id, targets, superrun_only=True):
    targets = strax.to_str_tuple(targets)
    plugins = self._get_plugins(targets=targets, run_id=run_id)
    configs = dict()
    for data_type, plugin in plugins.items():
        for k, v in plugin.config.items():
            if not isinstance(v, str):
                continue
            arg, extra_kwargs = straxen.URLConfig.split_url_kwargs(v)
            if "version" in extra_kwargs:
                configs[k] = (v, plugins[data_type])
    return configs



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventPeaks(strax.Plugin):
    """Add event number for peaks and drift times of all s2 depending on the largest s1.

    Link - https://xe1t-wiki.lngs.infn.it/doku.php?id=weiss:analysis:ms_plugin

    """

    __version__ = "0.0.2"
    depends_on = ("event_basics", "peak_basics", "peak_positions")
    provides = "peak_per_event"
    data_kind = "peaks"
    save_when = strax.SaveWhen.TARGET

    def infer_dtype(self):
        dtype = strax.time_fields + [
            ("drift_time", np.float32, "Drift time between main S1 and S2 in ns"),
            ("event_number", np.int64, "Event number in this dataset"),
        ]
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        split_peaks_ind = strax.fully_contained_in(peaks, events)
        result = np.zeros(len(peaks), self.dtype)
        strax.set_nan_defaults(result)

        # Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            result["drift_time"][split_peaks_ind == event_i] = (
                sp["center_time"] - event["s1_center_time"]
            )
        # Start of new part
        sorted_indices_split_peaks_ind = strax.stable_argsort(split_peaks_ind)
        mapping = {
            val: events["event_number"][i]
            for val, i in zip(
                np.unique(
                    split_peaks_ind[sorted_indices_split_peaks_ind][
                        split_peaks_ind[sorted_indices_split_peaks_ind] != -1
                    ]
                ),
                range(len(events["event_number"])),
            )
        }

        corrected_split_peaks_ind = np.array(
            [mapping[val] if val in mapping else val for val in split_peaks_ind]
        )

        result["event_number"] = corrected_split_peaks_ind
        # End of new part
        result["drift_time"][peaks["type"] != 2] = np.nan
        result["time"] = peaks["time"]
        result["endtime"] = strax.endtime(peaks)
        return result



import numpy as np
import numba
import strax
import straxen


export, __all__ = strax.exporter()


@export
class EventBasicsVanilla(strax.Plugin):
    """Computes the basic properties of the main/alternative S1/S2 within an event.

    The main S1 and alternative S1 are given by the largest two S1-Peaks within the event. The main
    S2 is given by the largest S2-Peak within the event, while alternative S2 is selected as the
    largest S2 other than main S2 in the time window [main S1 time, main S1 time + max drift time].

    """

    __version__ = "1.3.6"

    depends_on = ("events", "peak_basics", "peak_positions", "peak_proximity")
    provides = "event_basics"
    data_kind = "events"

    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )

    allow_posts2_s1s = straxen.URLConfig(
        default=False,
        infer_type=False,
        help="Allow S1s past the main S2 to become the main S1 and S2",
    )

    force_main_before_alt = straxen.URLConfig(
        default=False,
        infer_type=False,
        help="Make the alternate S1 (and likewise S2) the main S1 if occurs before the main S1.",
    )

    force_alt_s2_in_max_drift_time = straxen.URLConfig(
        default=True,
        infer_type=False,
        help="Make sure alt_s2 is in max drift time starting from main S1",
    )

    event_s1_min_coincidence = straxen.URLConfig(
        default=2,
        infer_type=False,
        help=(
            "Event level S1 min coincidence. Should be >= s1_min_coincidence "
            "in the peaklet classification"
        ),
    )

    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z,
        infer_type=False,
        help="Total length of the TPC from the bottom of gate to the top of cathode wires [cm]",
    )

    def infer_dtype(self):
        # Basic event properties
        self._set_posrec_save()
        self._set_dtype_requirements()
        dtype = strax.time_fields + [
            ("n_peaks", np.int32, "Number of peaks in the event"),
            ("drift_time", np.float32, "Drift time between main S1 and S2 in ns"),
            ("event_number", np.int64, "Event number in this dataset"),
        ]

        dtype += self._get_si_dtypes(self.peak_properties)

        dtype += [
            (
                f"s1_max_diff",
                np.int32,
                f"Main S1 largest time difference between apexes of hits [ns]",
            ),
            (
                f"alt_s1_max_diff",
                np.int32,
                f"Alternate S1 largest time difference between apexes of hits [ns]",
            ),
            (
                f"s1_min_diff",
                np.int32,
                f"Main S1 smallest time difference between apexes of hits [ns]",
            ),
            (
                f"alt_s1_min_diff",
                np.int32,
                f"Alternate S1 smallest time difference between apexes of hits [ns]",
            ),
            (
                f"s1_first_channel",
                np.int16,
                f"Main S1 first channel/PMT number (sorted by apexes of hits)",
            ),
            (
                f"alt_s1_first_channel",
                np.int16,
                f"Alternate S1 first channel/PMT number (sorted by apexes of hits)",
            ),
            (
                f"s1_last_channel",
                np.int16,
                f"Main S1 last channel/PMT number (sorted by apexes of hits)",
            ),
            (
                f"alt_s1_last_channel",
                np.int16,
                f"Alternate last channel/PMT number (sorted by apexes of hits)",
            ),
        ]

        dtype += [
            (f"s2_x", np.float32, f"Main S2 reconstructed X position, uncorrected [cm]"),
            (f"s2_y", np.float32, f"Main S2 reconstructed Y position, uncorrected [cm]"),
            (f"alt_s2_x", np.float32, f"Alternate S2 reconstructed X position, uncorrected [cm]"),
            (f"alt_s2_y", np.float32, f"Alternate S2 reconstructed Y position, uncorrected [cm]"),
            (f"area_before_main_s2", np.float32, f"Sum of areas before Main S2 [PE]"),
            (
                f"large_s2_before_main_s2_area",
                np.float32,
                f"The largest S2 before the Main S2 [PE]",
            ),
            (
                f"large_s2_before_main_s2_index",
                np.int32,
                f"Index of the largest S2 before the Main S2",
            ),
            (
                f"large_s2_before_main_s2_center_time",
                np.int64,
                f"Center time of the largest S2 before the Main S2 [ns]",
            ),
        ]

        dtype += self._get_posrec_dtypes()
        return dtype

    def _set_dtype_requirements(self):
        """Needs to be run before inferring dtype as it is needed there."""
        # Properties to store for each peak (main and alternate S1 and S2)
        self.peak_properties = (
            ("time", np.int64, "start time since unix epoch [ns]"),
            ("center_time", np.int64, "weighted average center time since unix epoch [ns]"),
            ("median_time", np.float32, "weighted relative median time of the peak [ns]"),
            ("endtime", np.int64, "end time since unix epoch [ns]"),
            ("area", np.float32, "area, uncorrected [PE]"),
            ("n_channels", np.int16, "count of contributing PMTs"),
            ("top_n_channels", np.int16, "count of contributing top PMTs"),
            ("n_hits", np.int32, "count of hits contributing at least one sample to the peak"),
            ("proximity_score", np.float32, "proximity score in (time, space)"),
            ("n_competing_left", np.int32, "number of competing peaks left of the main peak"),
            ("n_competing", np.int32, "number of competing peaks"),
            ("max_pmt", np.int16, "PMT number which contributes the most PE"),
            ("max_pmt_area", np.float32, "area in the largest-contributing PMT (PE)"),
            ("range_50p_area", np.float32, "width, 50% area [ns]"),
            ("range_90p_area", np.float32, "width, 90% area [ns]"),
            ("rise_time", np.float32, "time between 10% and 50% area quantiles [ns]"),
            ("area_fraction_top", np.float32, "fraction of area seen by the top PMT array"),
            ("tight_coincidence", np.int16, "channel within tight range of mean"),
            ("n_saturated_channels", np.int16, "total number of saturated channels"),
            ("merged", bool, "is merged from peaklets"),
        )

    def setup(self):
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)

    @staticmethod
    def _get_si_dtypes(peak_properties):
        """Get properties for S1/S2 from peaks directly."""
        si_dtype = []
        for s_i in [1, 2]:
            # Peak indices
            si_dtype += [
                (f"s{s_i}_index", np.int32, f"Main S{s_i} peak index in event"),
                (f"alt_s{s_i}_index", np.int32, f"Alternate S{s_i} peak index in event"),
            ]

            # Peak properties
            for name, dt, comment in peak_properties:
                si_dtype += [
                    (f"s{s_i}_{name}", dt, f"Main S{s_i} {comment}"),
                    (f"alt_s{s_i}_{name}", dt, f"Alternate S{s_i} {comment}"),
                ]

            # Drifts and delays
            si_dtype += [
                (
                    f"alt_s{s_i}_interaction_drift_time",
                    np.float32,
                    f"Drift time using alternate S{s_i} [ns]",
                ),
                (f"alt_s{s_i}_delay", np.int32, f"Time between main and alternate S{s_i} [ns]"),
            ]
        return si_dtype

    def _set_posrec_save(self):
        """Parse x_mlp et cetera if needed to get the algorithms used and set required class
        attributes."""
        posrec_fields = self.deps["peak_positions"].dtype_for("peak_positions").names
        posrec_names = [d.split("_")[-1] for d in posrec_fields if "x_" in d]

        # Preserve order. "set" is not ordered and dtypes should always be ordered
        self.pos_rec_labels = list(set(posrec_names))
        self.pos_rec_labels.sort()

        self.posrec_save = [(xy + alg) for xy in ["x_", "y_"] for alg in self.pos_rec_labels]

    def _get_posrec_dtypes(self):
        """Get S2 positions for each of the position reconstruction algorithms."""
        posrec_dtpye = []

        for alg in self.pos_rec_labels:
            # S2 positions
            posrec_dtpye += [
                (
                    f"s2_x_{alg}",
                    np.float32,
                    f"Main S2 {alg}-reconstructed X position, uncorrected [cm]",
                ),
                (
                    f"s2_y_{alg}",
                    np.float32,
                    f"Main S2 {alg}-reconstructed Y position, uncorrected [cm]",
                ),
                (
                    f"alt_s2_x_{alg}",
                    np.float32,
                    f"Alternate S2 {alg}-reconstructed X position, uncorrected [cm]",
                ),
                (
                    f"alt_s2_y_{alg}",
                    np.float32,
                    f"Alternate S2 {alg}-reconstructed Y position, uncorrected [cm]",
                ),
            ]

        return posrec_dtpye

    def compute(self, events, peaks):
        result = np.zeros(len(events), dtype=self.dtype)
        strax.set_nan_defaults(result)

        split_peaks = strax.split_by_containment(peaks, events)

        result["time"] = events["time"]
        result["endtime"] = events["endtime"]
        result["event_number"] = events["event_number"]

        self.fill_events(result, split_peaks)
        return result

    # If copy_largest_peaks_into_event is ever numbafied, also numbafy this function
    def fill_events(self, result_buffer, split_peaks):
        """Loop over the events and peaks within that event."""
        for event_i, peaks_in_event_i in enumerate(split_peaks):
            n_peaks = len(peaks_in_event_i)
            result_buffer[event_i]["n_peaks"] = n_peaks

            if not n_peaks:
                raise ValueError(f"No peaks within event {event_i}?")

            self.fill_result_i(result_buffer[event_i], peaks_in_event_i)

    def fill_result_i(self, event, peaks):
        """For a single event with the result_buffer."""
        # Consider S2s first, then S1s (to enable allow_posts2_s1s = False)
        # number_of_peaks=0 selects all available s2 and sort by area
        largest_s2s, s2_idx = self.get_largest_sx_peaks(peaks, s_i=2, number_of_peaks=0)

        if not self.allow_posts2_s1s and len(largest_s2s):
            s1_latest_time = largest_s2s[0]["time"]
        else:
            s1_latest_time = np.inf

        largest_s1s, s1_idx = self.get_largest_sx_peaks(
            peaks,
            s_i=1,
            s1_before_time=s1_latest_time,
            s1_min_coincidence=self.event_s1_min_coincidence,
        )

        if self.force_alt_s2_in_max_drift_time:
            s2_idx, largest_s2s = self.find_main_alt_s2(
                largest_s1s,
                s2_idx,
                largest_s2s,
                self.drift_time_max,
            )
        else:
            # Select only the largest two S2s
            largest_s2s, s2_idx = largest_s2s[0:2], s2_idx[0:2]

        if self.force_main_before_alt:
            s2_order = strax.stable_argsort(largest_s2s["time"])
            largest_s2s = largest_s2s[s2_order]
            s2_idx = s2_idx[s2_order]

        self.set_sx_index(event, s1_idx, s2_idx)
        self.set_event_properties(event, largest_s1s, largest_s2s, peaks)

        # Loop over S1s and S2s and over main / alt.
        for s_i, largest_s_i in enumerate([largest_s1s, largest_s2s], 1):
            # Largest index 0 -> main sx, 1 -> alt sx
            for largest_index, main_or_alt in enumerate(["s", "alt_s"]):
                peak_properties_to_save = [name for name, _, _ in self.peak_properties]
                if s_i == 1:
                    peak_properties_to_save += [
                        "max_diff",
                        "min_diff",
                        "first_channel",
                        "last_channel",
                    ]
                elif s_i == 2:
                    peak_properties_to_save += ["x", "y"]
                    peak_properties_to_save += self.posrec_save
                field_names = [f"{main_or_alt}{s_i}_{name}" for name in peak_properties_to_save]
                self.copy_largest_peaks_into_event(
                    event, largest_s_i, largest_index, field_names, peak_properties_to_save
                )

    @staticmethod
    @numba.njit
    def find_main_alt_s2(largest_s1s, s2_idx, largest_s2s, drift_time_max):
        """Require alt_s2 happens between main S1 and maximum drift time."""
        if len(largest_s1s) > 0 and len(largest_s2s) > 1:
            # If there is a valid s1-s2 pair and has a second s2, then check alt s2 validity
            s2_after_s1 = largest_s2s["center_time"] > largest_s1s[0]["center_time"]
            s2_before_max_drift_time = (
                largest_s2s["center_time"] - largest_s1s[0]["center_time"]
            ) < 1.01 * drift_time_max
            mask = s2_after_s1 & s2_before_max_drift_time
            # The selection avoids main_S2
            mask[0] = True
            # Take main and the largest valid alt_S2
            s2_idx, largest_s2s = s2_idx[mask], largest_s2s[mask]
        return s2_idx[:2], largest_s2s[:2]

    @staticmethod
    @numba.njit
    def set_event_properties(result, largest_s1s, largest_s2s, peaks):
        """Get properties like drift time and area before main S2."""
        # Compute drift times only if we have a valid S1-S2 pair
        if len(largest_s1s) > 0 and len(largest_s2s) > 0:
            result["drift_time"] = largest_s2s[0]["center_time"] - largest_s1s[0]["center_time"]
            if len(largest_s1s) > 1:
                result["alt_s1_interaction_drift_time"] = (
                    largest_s2s[0]["center_time"] - largest_s1s[1]["center_time"]
                )
                result["alt_s1_delay"] = (
                    largest_s1s[1]["center_time"] - largest_s1s[0]["center_time"]
                )
            if len(largest_s2s) > 1:
                result["alt_s2_interaction_drift_time"] = (
                    largest_s2s[1]["center_time"] - largest_s1s[0]["center_time"]
                )
                result["alt_s2_delay"] = (
                    largest_s2s[1]["center_time"] - largest_s2s[0]["center_time"]
                )

        # areas before main S2
        if len(largest_s2s):
            peaks_before_ms2 = ~np.isnan(peaks["area"])
            peaks_before_ms2 &= peaks["center_time"] < largest_s2s[0]["center_time"]
            result["area_before_main_s2"] = np.sum(peaks["area"][peaks_before_ms2])

            s2_before_ms2 = peaks_before_ms2 & (peaks["type"] == 2)
            if np.any(s2_before_ms2):
                i = np.arange(len(peaks))[s2_before_ms2][
                    np.argmax(peaks["area"][s2_before_ms2]).item()
                ]
                result["large_s2_before_main_s2_area"] = peaks["area"][i]
                result["large_s2_before_main_s2_index"] = i
                result["large_s2_before_main_s2_center_time"] = peaks["center_time"][i]
        return result

    @staticmethod
    # @numba.njit <- works but slows if fill_events is not numbafied
    def get_largest_sx_peaks(
        peaks, s_i, s1_before_time=np.inf, s1_min_coincidence=0, number_of_peaks=2
    ):
        """Get the largest S1/S2.

        For S1s allow a min coincidence and max time

        """
        # Find all peaks of this type (S1 or S2)
        s_mask = peaks["type"] == s_i
        if s_i == 1:
            s_mask &= peaks["time"] < s1_before_time
            s_mask &= peaks["tight_coincidence"] >= s1_min_coincidence

        selected_peaks = peaks[s_mask]
        s_index = np.arange(len(peaks))[s_mask]
        largest_peaks = strax.stable_argsort(selected_peaks["area"])[-number_of_peaks:][::-1]
        return selected_peaks[largest_peaks], s_index[largest_peaks]

    # If only we could numbafy this... Unfortunatly we cannot.
    # Perhaps we could one day consider doing something like strax.copy_to_buffer
    @staticmethod
    def copy_largest_peaks_into_event(
        result,
        largest_s_i,
        main_or_alt_index,
        result_fields,
        peak_fields,
    ):
        """For one event, write all the peak_fields (e.g. "area") of the peak (largest_s_i) into
        their associated field in the event (e.g. s1_area), main_or_alt_index differentiates between
        main (index 0) and alt (index 1)"""
        index_not_in_list_of_largest_peaks = main_or_alt_index >= len(largest_s_i)
        if index_not_in_list_of_largest_peaks:
            # There is no such peak. E.g. main_or_alt_index == 1 but largest_s_i = ["Main S1"]
            # Asking for index 1 doesn't work on a len 1 list of peaks.
            return

        for i, ev_field in enumerate(result_fields):
            p_field = peak_fields[i]
            if p_field not in ev_field:
                raise ValueError("Event fields must derive from the peak fields")
            result[ev_field] = largest_s_i[main_or_alt_index][p_field]

    @staticmethod
    # @numba.njit <- works but slows if fill_events is not numbafied
    def set_sx_index(res, s1_idx, s2_idx):
        if len(s1_idx):
            res["s1_index"] = s1_idx[0]
            if len(s1_idx) > 1:
                res["alt_s1_index"] = s1_idx[1]
        if len(s2_idx):
            res["s2_index"] = s2_idx[0]
            if len(s2_idx) > 1:
                res["alt_s2_index"] = s2_idx[1]



import strax
import straxen
import numpy as np

export, __all__ = strax.exporter()


@export
class EventInfoMS(strax.Plugin):
    """Plugin to collect multiple-scatter event observables."""

    __version__ = "0.0.2"
    depends_on = (
        "event_info",
        "peak_basics",
        "peak_per_event",
        "peak_corrections",
        "peak_positions",
    )
    provides = "event_ms_naive"
    save_when = strax.SaveWhen.TARGET

    # config options don't double cache things from the resource cache!
    g1 = straxen.URLConfig(
        default="bodega://g1?bodega_version=v2",
        help="S1 gain in PE / photons produced",
    )
    g2 = straxen.URLConfig(
        default="bodega://g2?bodega_version=v2",
        help="S2 gain in PE / electrons produced",
    )
    lxe_w = straxen.URLConfig(default=13.7e-3, help="LXe work function in quanta/keV")
    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )
    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z,
        infer_type=False,
        help="Total length of the TPC from the bottom of gate to the top of cathode wires [cm]",
    )

    ms_window_fac = straxen.URLConfig(
        default=1.01,
        type=(int, float),
        help="Max drift time window to look for peaks in multiple scatter events",
    )

    def infer_dtype(self):
        dtype = strax.time_fields + [
            (("Sum of S1 areas in event", "s1_sum"), np.float32),
            (
                ("Corrected S1 area based on average position of S2s in event", "cs1_multi"),
                np.float32,
            ),
            (
                (
                    (
                        "Corrected S1 area based on average position of S2s in event before"
                        " time-dep LY correction"
                    ),
                    "cs1_multi_wo_timecorr",
                ),
                np.float32,
            ),
            (("Sum of S2 areas in event", "s2_sum"), np.float32),
            (("Sum of corrected S2 areas in event", "cs2_sum"), np.float32),
            (
                (
                    "Sum of corrected S2 areas in event S2 before elife correction",
                    "cs2_wo_timecorr_sum",
                ),
                np.float32,
            ),
            (
                (
                    "Sum of corrected S2 areas in event before SEG/EE and elife corrections",
                    "cs2_wo_elifecorr_sum",
                ),
                np.float32,
            ),
            (("Average of S2 area fraction top in event", "cs2_area_fraction_top_avg"), np.float32),
            (("Sum of the energy estimates in event", "ces_sum"), np.float32),
            (("Sum of the charge estimates in event", "e_charge_sum"), np.float32),
            (("Average x position of S2s in event", "x_avg"), np.float32),
            (("Average y position of S2s in event", "y_avg"), np.float32),
            (("Average observed z position of energy deposits in event", "z_obs_avg"), np.float32),
            (("Number of S2s in event", "multiplicity"), np.int32),
        ]
        return dtype

    def setup(self):
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)

    def cs1_to_e(self, x):
        return self.lxe_w * x / self.g1

    def cs2_to_e(self, x):
        return self.lxe_w * x / self.g2

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        result = np.zeros(len(events), self.infer_dtype())

        # Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            cond = (sp["type"] == 2) & (sp["drift_time"] > 0)
            cond &= (sp["drift_time"] < self.ms_window_fac * self.drift_time_max) & (sp["cs2"] > 0)
            result["s2_sum"][event_i] = np.nansum(sp[cond]["area"])
            result["cs2_sum"][event_i] = np.nansum(sp[cond]["cs2"])
            result["cs2_wo_timecorr_sum"][event_i] = np.nansum(sp[cond]["cs2_wo_timecorr"])
            result["cs2_wo_elifecorr_sum"][event_i] = np.nansum(sp[cond]["cs2_wo_elifecorr"])
            result["s1_sum"][event_i] = np.nansum(sp[sp["type"] == 1]["area"])

            if np.sum(sp[cond]["cs2"]) > 0:
                result["cs1_multi_wo_timecorr"][event_i] = event["s1_area"] * np.average(
                    sp[cond]["s1_xyz_correction_factor"], weights=sp[cond]["cs2"]
                )
                result["cs1_multi"][event_i] = result["cs1_multi_wo_timecorr"][
                    event_i
                ] * np.average(
                    sp[cond]["s1_rel_light_yield_correction_factor"], weights=sp[cond]["cs2"]
                )
                result["x_avg"][event_i] = np.average(sp[cond]["x"], weights=sp[cond]["cs2"])
                result["y_avg"][event_i] = np.average(sp[cond]["y"], weights=sp[cond]["cs2"])
                result["z_obs_avg"][event_i] = np.average(
                    sp[cond]["z_obs_ms"], weights=sp[cond]["cs2"]
                )
                result["cs2_area_fraction_top_avg"][event_i] = np.average(
                    sp[cond]["cs2_area_fraction_top"], weights=sp[cond]["cs2"]
                )
                result["multiplicity"][event_i] = len(sp[cond]["area"])

        el = self.cs1_to_e(result["cs1_multi"])
        ec = self.cs2_to_e(result["cs2_sum"])
        result["ces_sum"] = el + ec
        result["e_charge_sum"] = ec
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)
        return result



import os
import json
from warnings import warn
from typing import Dict

import numpy as np
import strax
from utilix import xent_collection

try:
    import admix
    from rucio.common.exception import DataIdentifierNotFound

    HAVE_ADMIX = True
except (ImportError, AttributeError):
    HAVE_ADMIX = False

export, __all__ = strax.exporter()

__all__.extend(["HAVE_ADMIX"])


@export
class TooMuchDataError(Exception):
    pass


@export
class RucioRemoteFrontend(strax.StorageFrontend):
    """Uses the rucio client for the data find."""

    storage_type = strax.StorageType.REMOTE
    local_did_cache = None
    path = None

    def __init__(
        self,
        staging_dir="./strax_data",
        rses_only=tuple(),
        download_heavy=False,
        remove_heavy=False,
        tries=3,
        num_threads=1,
        stage=False,
        *args,
        **kwargs,
    ):
        """
        :param download_heavy: option to allow downloading of heavy data through RucioRemoteBackend
        :param remove_heavy: option to remove heavy data from the RucioRemoteBackend after reading
        :param args: Passed to strax.StorageFrontend
        :param kwargs: Passed to strax.StorageFrontend
        :param rses_only: tuple, limits RSE selection to these options if provided
        """
        super().__init__(*args, **kwargs)
        self.readonly = True
        self.collection = xent_collection()
        self.backends = []

        if HAVE_ADMIX:
            self.backends = [
                RucioRemoteBackend(
                    staging_dir=staging_dir,
                    rses_only=rses_only,
                    download_heavy=download_heavy,
                    remove_heavy=remove_heavy,
                    tries=tries,
                    num_threads=num_threads,
                    stage=stage,
                ),
            ]
        else:
            self.log.warning(
                "You passed use_remote=True to rucio fronted, "
                "but you don't have access to admix/rucio! Using local backend only."
            )

    def find_several(self, keys, **kwargs):
        # for performance, dont do find_several with this storage frontend
        # we basically do the same query we would do in the RunDB plugin
        return np.zeros_like(keys, dtype=bool).tolist()

    def _find(self, key: strax.DataKey, write, allow_incomplete, fuzzy_for, fuzzy_for_options):
        did = key_to_rucio_did(key)
        if allow_incomplete or write:
            raise RuntimeError(
                "Allow incomplete/writing is not allowed for "
                f"{self.__class.__name__} since data might not be "
                "continuous"
            )
        try:
            for b in self.backends:
                rse = b._get_rse(did, state="OK")
                if rse:
                    return "RucioRemoteBackend", did
        except DataIdentifierNotFound:
            pass

        raise strax.DataNotAvailable

    def find(self, key: strax.DataKey, write=False, check_broken=False, **kwargs):
        # Overwrite defaults of super().find()
        return super().find(key, write, check_broken, **kwargs)


@export
class RucioRemoteBackend(strax.FileSytemBackend):
    """Get data from remote Rucio RSE."""

    # datatypes we don't want to download since they're too heavy
    heavy_types = ["raw_records", "raw_records_nv", "raw_records_he"]

    # for caching RSE locations
    dset_cache: Dict[str, str] = {}

    def __init__(
        self,
        staging_dir,
        rses_only=tuple(),
        download_heavy=False,
        remove_heavy=False,
        tries=3,
        num_threads=1,
        stage=False,
        **kwargs,
    ):
        """
        :param staging_dir: Path (a string) where to save data. Must be
            a writable location.
        :param download_heavy: Whether or not to allow downloads of the
            heaviest data (raw_records*, less aqmon and MV)
        :param remove_heavy: Whether or not to remove the heaviest data after reading
        :param kwargs: Passed to strax.FileSystemBackend
        :param rses_only: tuple, limits RSE selection to these options if provided
        """
        mess = (
            f"You told the rucio backend to download data to {staging_dir}, "
            "but that path is not writable by your user"
        )
        if os.path.exists(staging_dir):
            if not os.access(staging_dir, os.W_OK):
                raise PermissionError(mess)
        else:
            try:
                os.makedirs(staging_dir)
            except OSError:
                raise PermissionError(mess)
        super().__init__(**kwargs)
        self.staging_dir = staging_dir
        self.rses_only = strax.to_str_tuple(rses_only)
        self.download_heavy = download_heavy
        self.remove_heavy = remove_heavy
        self.tries = tries
        self.num_threads = num_threads
        self.stage = stage

    def _get_rse(self, dset_did, **filters):
        """Determine the appropriate Rucio Storage Element (RSE) for a dataset.

        :param dset_did (str) :The dataset identifier.
        :return (str) : The selected RSEs.
        ------
        Uses self.rses_only to filter available RSEs if set.

        """
        rses = admix.rucio.get_rses(dset_did, **filters)
        rses = list(set(rses) & set(self.rses_only)) if self.rses_only else rses
        rse = admix.downloader.determine_rse(rses)
        return rse

    def _get_metadata(self, dset_did, **kwargs):
        if dset_did in self.dset_cache:
            rse = self.dset_cache[dset_did]
        else:
            rse = self._get_rse(dset_did)
            self.dset_cache[dset_did] = rse

        metadata_did = strax.RUN_METADATA_PATTERN % dset_did
        warn(f"Downloading {metadata_did} from {rse}")
        downloaded = admix.download(
            metadata_did,
            location=self.staging_dir,
            tries=self.tries,
            num_threads=self.num_threads,
            rse=rse,
            stage=self.stage,
        )
        if len(downloaded) != 1:
            raise ValueError(f"{metadata_did} should be a single file. We found {len(downloaded)}.")
        metadata_path = downloaded[0]
        # check again
        if not os.path.exists(metadata_path):
            raise FileNotFoundError(f"No metadata found at {metadata_path}")

        with open(metadata_path, mode="r") as f:
            return json.loads(f.read())

    def _read_chunk(self, dset_did, chunk_info, dtype, compressor):
        base_dir = os.path.join(self.staging_dir, did_to_dirname(dset_did))
        chunk_file = chunk_info["filename"]
        chunk_path = os.path.abspath(os.path.join(base_dir, chunk_file))
        number, datatype, hsh = parse_rucio_did(dset_did)
        if not os.path.exists(chunk_path):
            if datatype in self.heavy_types and not self.download_heavy:
                error_msg = (
                    "For space reasons we don't want to have everyone "
                    "downloading raw data. If you know what you're "
                    "doing, pass download_heavy=True to the Rucio "
                    "frontend. If not, check your context and/or ask "
                    "someone if this raw data is needed locally."
                )
                warn(error_msg)
                raise strax.DataNotAvailable
            scope, name = dset_did.split(":")
            if dset_did in self.dset_cache:
                rse = self.dset_cache[dset_did]
            else:
                rse = self._get_rse(dset_did)
                self.dset_cache[dset_did] = rse

            chunk_did = f"{scope}:{chunk_file}"
            warn(f"Downloading {chunk_did} from {rse}")
            downloaded = admix.download(
                chunk_did,
                location=self.staging_dir,
                tries=self.tries,
                num_threads=self.num_threads,
                rse=rse,
                stage=self.stage,
            )
            if len(downloaded) != 1:
                raise ValueError(
                    f"{chunk_did} should be a single file. We found {len(downloaded)}."
                )
            assert chunk_path == downloaded[0]

        # check again
        if not os.path.exists(chunk_path):
            raise FileNotFoundError(f"No chunk file found at {chunk_path}")

        data = strax.load_file(chunk_path, dtype=dtype, compressor=compressor)

        if self.remove_heavy and datatype in self.heavy_types:
            warn(
                f"Removing {chunk_path} after reading since it's heavy data. "
                "This is a one-time operation."
            )
            os.remove(chunk_path)
        return data

    def _saver(self, dirname, metadata, **kwargs):
        raise NotImplementedError(
            "Cannot save directly into rucio (yet), upload with admix instead"
        )


@export
class RucioSaver(strax.Saver):
    """TODO: Saves data to rucio if you are the production user."""

    def __init__(self, *args, **kwargs):
        raise NotImplementedError


@export
def parse_rucio_did(did: str) -> tuple:
    """Parses a Rucio DID and returns a tuple of (number:int, dtype:str, hash:str)"""
    scope, name = did.split(":")
    number = int(scope.split("_")[1])
    dtype, hsh = name.split("-")
    return number, dtype, hsh


def did_to_dirname(did: str):
    """Takes a Rucio dataset DID and returns a dirname like used by strax.FileSystemBackend."""
    # make sure it's a DATASET did, not e.g. a FILE
    if len(did.split("-")) != 2:
        raise RuntimeError(
            f"The DID {did} does not seem to be a dataset DID. "
            "Is it possible you passed a file DID?"
        )
    dirname = did.replace(":", "-").replace("xnt_", "")
    return dirname


@export
def key_to_rucio_did(key: strax.DataKey) -> str:
    """Convert a strax.datakey to a rucio did field in rundoc."""
    return f"xnt_{key.run_id}:{key.data_type}-{key.lineage_hash}"



import strax
import straxen
from straxen.plugins.events._event_s2_positions_base import EventS2PositionBase


export, __all__ = strax.exporter()


@export
class EventS2PositionMLP(EventS2PositionBase):
    """MLP neural net for S2 position reconstruction at event level."""

    algorithm = "mlp"
    provides = "event_s2_positions_mlp"

    tf_model_mlp = straxen.URLConfig(
        default="tf://resource://xnt_mlp_SR0_mix_2000031_2000021_20211211.keras?fmt=abs_path",
        help=(
            'MLP model. Should be opened using the "tf" descriptor. '
            'Set to "None" to skip computation'
        ),
        cache=3,
    )



import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class Events(strax.OverlapWindowPlugin):
    """Plugin which defines an "event" in our TPC.

    An event is defined by peak(s) in fixed range of time around a peak
    which satisfies certain conditions:
        1. The triggering peak must have a certain area.
        2. The triggering peak must have less than "trigger_max_proximity".

    Note:
        The time range which defines an event gets chopped at the chunk
        boundaries. This happens at invalid boundaries of the

    """

    __version__ = "0.1.1"

    depends_on = ("peak_basics", "peak_proximity")
    provides = "events"
    data_kind = "events"

    save_when = strax.SaveWhen.EXPLICIT

    dtype = [
        ("event_number", np.int64, "Event number in this dataset"),
        ("time", np.int64, "Event start time in ns since the unix epoch"),
        ("endtime", np.int64, "Event end time in ns since the unix epoch"),
    ]

    events_seen = 0

    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )

    trigger_min_area = straxen.URLConfig(
        default=120,
        type=(int, float),
        help="Peaks must have more area (PE) than this to cause events",
    )

    trigger_feature = straxen.URLConfig(
        default="proximity_score",
        type=str,
        help="Feature as event building criterion",
    )

    trigger_max_proximity = straxen.URLConfig(
        default=1.5e-3,
        type=(int, float),
        help="Peaks must have less proximity score to cause events",
    )

    # TODO: extand the padding to cover all multiple scattering
    left_event_extension = straxen.URLConfig(
        default=int(0.25e6),
        type=(int, float),
        help=(
            "Extend events this many ns to the left from each "
            "triggering peak. This extension is added to the maximum "
            "drift time."
        ),
    )

    right_event_extension = straxen.URLConfig(
        default=int(0.25e6),
        type=(int, float),
        help="Extend events this many ns to the right from each triggering peak.",
    )

    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z,
        type=(int, float),
        help="Total length of the TPC from the bottom of gate to the top of cathode wires [cm]",
    )

    event_s1_min_coincidence = straxen.URLConfig(
        default=2,
        infer_type=False,
        help=(
            "Event level S1 min coincidence. Should be >= "
            "s1_min_coincidence in the peaklet classification"
        ),
    )

    s1_min_coincidence = straxen.URLConfig(
        default=2, type=int, help="Minimum tight coincidence necessary to make an S1"
    )

    event_exclude_s3 = straxen.URLConfig(
        default=True,
        type=bool,
        help="Event will not be built if S3 is present in the window",
    )

    diagnose_overlapping = straxen.URLConfig(
        track=False, default=True, infer_type=False, help="Enable runtime checks for disjointness"
    )

    def setup(self):
        if self.s1_min_coincidence > self.event_s1_min_coincidence:
            raise ValueError(
                "Peak s1 coincidence requirement should be smaller "
                "or equal to event_s1_min_coincidence"
            )
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)
        # Left_extension and right_extension should be computed in setup to be
        # reflected in cutax too.
        self.left_extension = self.left_event_extension + self.drift_time_max
        self.right_extension = self.right_event_extension

    def get_window_size(self):
        # Take a large window for safety, events can have long tails
        return 10 * (self.left_event_extension + self.drift_time_max + self.right_event_extension)

    def _is_triggering(self, peaks):
        _is_triggering = peaks["area"] > self.trigger_min_area
        _is_triggering &= peaks["type"] == 2
        _is_triggering &= peaks[self.trigger_feature] <= self.trigger_max_proximity
        return _is_triggering

    def compute(self, peaks, start, end):
        _is_triggering = self._is_triggering(peaks)

        triggers = peaks[_is_triggering]

        # Join nearby triggers
        t0, t1 = strax.find_peak_groups(
            triggers,
            gap_threshold=self.left_extension + self.right_extension + 1,
            left_extension=self.left_extension,
            right_extension=self.right_extension,
        )

        # Don't extend beyond the chunk boundaries
        # This will often happen for events near the invalid boundary of the
        # overlap processing (which should be thrown away)
        # The different chunking causes different results, but this is
        # NOT tracked by lineage
        t0 = np.clip(t0, start, end)
        t1 = np.clip(t1, start, end)

        result = np.zeros(len(t0), self.dtype)
        result["time"] = t0
        result["endtime"] = t1

        if self.event_exclude_s3:
            # type 3 high energy gas scintillation xenon:xenonnt:lsanchez:som_sr1b_summary_note
            split_peaks = strax.split_by_containment(peaks, result)
            has_s3 = np.array([np.any(sp["type"] == 3) for sp in split_peaks]).astype(bool)
            result = result[~has_s3]

        # because this is an overlap window plugin, event_number is not continuous
        result["event_number"] = np.arange(len(result)) + self.events_seen
        self.events_seen += len(result)

        if not result.size > 0:
            print("Found chunk without events?!")

        if self.diagnose_overlapping and len(result):
            # Check if the event windows overlap
            _event_window_do_not_overlap = (strax.endtime(result)[:-1] - result["time"][1:]) <= 0
            assert np.all(_event_window_do_not_overlap), "Events not disjoint"

        return result



from typing import Tuple, Union
import numpy as np
import strax
import straxen

from straxen.plugins.defaults import FAR_XYPOS_S2_TYPE

export, __all__ = strax.exporter()


@export
class PeaksVanilla(strax.Plugin):
    """Merge peaklets and merged S2s such that we obtain our peaks (replacing all peaklets that were
    later re-merged as S2s).

    As this step is computationally trivial, never save this plugin.

    """

    __version__ = "0.1.2"

    depends_on: Union[Tuple[str, ...], str] = (
        "peaklets",
        "enhanced_peaklet_classification",
        "merged_s2s",
    )
    data_kind = "peaks"
    provides = "peaks"
    compressor = "zstd"
    save_when = strax.SaveWhen.EXPLICIT

    diagnose_sorting = straxen.URLConfig(
        track=False,
        default=False,
        infer_type=False,
        help="Enable runtime checks for sorting and disjointness",
    )

    merge_s0 = straxen.URLConfig(
        default=True,
        type=bool,
        help="Merge S0s into merged S2s",
    )

    def infer_dtype(self):
        # In case enhanced_peaklet_classification has more fields than peaklets,
        # we need to merge them
        peaklets_dtype = self.deps["peaklets"].dtype_for("peaklets")
        peaklet_classification_dtype = self.deps["enhanced_peaklet_classification"].dtype_for(
            "enhanced_peaklet_classification"
        )
        merged_dtype = strax.merged_dtype((peaklets_dtype, peaklet_classification_dtype))
        return merged_dtype

    def compute(self, peaklets, merged_s2s):
        _peaks = self.replace_merged(peaklets, merged_s2s, merge_s0=self.merge_s0)

        if self.diagnose_sorting:
            assert np.all(np.diff(_peaks["time"]) >= 0), "Peaks not sorted"
            to_check = _peaks["type"] == 2

            assert np.all(
                _peaks["time"][to_check][1:] >= strax.endtime(_peaks)[to_check][:-1]
            ), "Peaks not disjoint"

        peaks = np.zeros(len(_peaks), dtype=self.dtype)
        strax.copy_to_buffer(_peaks, peaks, f"_copy_requested_{self.provides[0]}_fields")
        return peaks

    @staticmethod
    def replace_merged(peaklets, merged_s2s, merge_s0=True):
        # pick out type FAR_XYPOS_S2_TYPE because they might overlap with other merged S2s
        peaklets_unmerged = peaklets["type"] == FAR_XYPOS_S2_TYPE
        # if not S0 not merged, also pick out type 0
        if merge_s0:
            # of course in all cases we will pick out type 1
            peaklets_unmerged |= peaklets["type"] == 1
        else:
            peaklets_unmerged |= np.isin(peaklets["type"], [0, 1])
        peaks = strax.replace_merged(peaklets[~peaklets_unmerged], merged_s2s)
        peaks = strax.sort_by_time(np.concatenate([peaklets[peaklets_unmerged], peaks]))
        return peaks



"""Core functions of the DAQ, mostly used in straxen/bin."""

import pymongo
import utilix
from utilix import uconfig
import straxen
from datetime import datetime, timedelta
import pytz

ceph_folder = "/live_data/xenonnt/"
output_folder = "/data/xenonnt_processed/"
pre_folder = "/data/pre_processed/"
non_registered_folder = "/data/xenonnt_unregistered/"


class DataBases:
    def __init__(self, production=False):
        self.production = production
        # DAQ database
        daq_db_name = "daq"
        daq_uri = straxen.get_mongo_uri(
            header="rundb_admin",
            user_key="mongo_daq_username",
            pwd_key="mongo_daq_password",
            url_key="mongo_daq_url",
        )
        daq_client = pymongo.MongoClient(daq_uri)
        self.daq_db = daq_client[daq_db_name]
        self.bs_coll = self.daq_db["eb_monitor"]
        self.ag_stat_coll = self.daq_db["aggregate_status"]
        self.log_coll = self.daq_db["log"]

        # Runs database
        run_dbname = straxen.uconfig.get("rundb_admin", "mongo_rdb_database")
        run_collname = "runs"
        if production:
            self.run_db = self.get_admin_client()[run_dbname]
        else:
            # Please note, this is a read only account on the rundb
            run_uri = straxen.get_mongo_uri()
            run_client = pymongo.MongoClient(run_uri)
            self.run_db = run_client[run_dbname]
        self.run_coll = self.run_db[run_collname]

    @staticmethod
    def get_admin_client():
        # We want admin access to start writing data!
        mongo_url = uconfig.get("rundb_admin", "mongo_rdb_url")
        mongo_user = uconfig.get("rundb_admin", "mongo_rdb_username")
        mongo_password = uconfig.get("rundb_admin", "mongo_rdb_password")
        mongo_database = uconfig.get("rundb_admin", "mongo_rdb_database")

        collection = utilix.rundb.xent_collection(
            url=mongo_url, user=mongo_user, password=mongo_password, database=mongo_database
        )

        # Do not delete the client!
        return collection.database.client

    def log_warning(
        self,
        message,
        priority="warning",
        run_id=None,
        production=True,
        user="daq_process",
    ):
        """Report a warning to the terminal (using the logging module) and the DAQ log DB.

        :param message: insert string into log_coll
        :param priority: severity of warning. Can be:
            info: 1,
            warning: 2,
            <any other valid python logging level, e.g. error or fatal>: 3
        :param run_id: optional run id.

        """
        if not production:
            return

        # Log according to redax rules
        # https://github.com/coderdj/redax/blob/master/MongoLog.hh#L22
        warning_message = {
            "message": message,
            "user": user,
            "priority": dict(
                debug=0,
                info=1,
                warning=2,
                error=3,
                fatal=4,
            ).get(priority.lower(), 3),
        }
        if run_id is not None:
            warning_message.update({"runid": int(run_id)})
        self.log_coll.insert_one(warning_message)


def now(plus=0):
    """Now in utc time."""
    return datetime.now(pytz.utc) + timedelta(seconds=plus)



"""S1 Position reconstruction for Xenon-nT
Author and maintainer: Matteo Guida (guidam@mpi-hd.mpg.de)"""

from typing import Optional

import strax
import straxen
import numpy as np
from warnings import warn

export, __all__ = strax.exporter()


@export
class PeakS1PositionBase(strax.Plugin):
    """Base pluging for S1 position reconstruction at peak level.

    Provides x_algorithm, y_algorithm, z_algorithm for all peaks > than min-reconstruction area
    based on both the top and bottom array.

    """

    __version__ = "0.0.0"
    depends_on = "peaks"

    algorithm: Optional[str] = None
    compressor = "zstd"

    min_s1_area_s1_posrec = straxen.URLConfig(
        help="Skip reconstruction if area (PE) is less than this", default=1000, infer_type=False
    )

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(
                f"Base class should not be used without algorithm as done in {__class__.__name__}"
            )
        dtype = [
            (
                "x_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S1 X position (cm), uncorrected",
            ),
            (
                "y_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S1 Y position (cm), uncorrected",
            ),
            (
                "z_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S1 Z position (cm), uncorrected",
            ),
        ]

        dtype += strax.time_fields
        return dtype

    def get_tf_model(self):
        """Simple wrapper to have several tf_model_s1_cnn, ..

        point to this same function in the compute method

        """
        model = getattr(self, f"tf_model_{self.algorithm}", None)
        if model is None:
            warn(
                f"Setting model to None for {self.__class__.__name__} will "
                f"set only nans as output for {self.algorithm}"
            )
        if isinstance(model, str):
            raise ValueError(
                f"open files from tf:// protocol! Got {model} "
                "instead, see tests/test_s1_posrec.py for examples."
            )
        return model

    def compute(self, peaks):
        result = np.ones(len(peaks), dtype=self.dtype)
        result["time"], result["endtime"] = peaks["time"], strax.endtime(peaks)

        result["x_" + self.algorithm] *= np.nan
        result["y_" + self.algorithm] *= np.nan
        result["z_" + self.algorithm] *= np.nan

        model = self.get_tf_model()

        # Reconstruct position only for large peaks, otherwise severe inaccuracy.
        peak_mask = peaks["area_per_channel"].sum(axis=1) > self.min_s1_area_s1_posrec
        peak_mask &= peaks["type"] < 2

        if not np.sum(peak_mask):
            # No peaks fulfilling the conditions, return nan array.
            return result

        _in = peaks["area_per_channel"][peak_mask]

        with np.errstate(divide="ignore", invalid="ignore"):
            # Normalise patters by dividing by largest PMT output between the two arrays.
            _in = _in / _in.max(axis=1, keepdims=True)

        # Getting actual position reconstruction
        _out = model.predict(_in)

        # writing output to the result
        result["x_" + self.algorithm][peak_mask] = _out[:, 0]
        result["y_" + self.algorithm][peak_mask] = _out[:, 1]
        result["z_" + self.algorithm][peak_mask] = _out[:, 2]
        return result



from typing import Union

import numpy as np
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakletClassificationVanilla(strax.Plugin):
    """Classify peaklets as unknown, S1, or S2."""

    __version__ = "3.0.4"

    depends_on = "peaklets"
    provides: Union[str, tuple] = "peaklet_classification"
    dtype = strax.peak_interval_dtype + [("type", np.int8, "Classification of the peak(let)")]

    s1_risetime_area_parameters = straxen.URLConfig(
        default=(50, 80, 12),
        type=(list, tuple),
        help="norm, const, tau in the empirical boundary in the risetime-area plot",
    )

    s1_risetime_aft_parameters = straxen.URLConfig(
        default=(-1, 2.6),
        type=(list, tuple),
        help=(
            "Slope and offset in exponential of emperical boundary in the rise time-AFT "
            "plot. Specified as (slope, offset)"
        ),
    )

    s1_flatten_threshold_aft = straxen.URLConfig(
        default=(0.6, 100),
        type=(tuple, list),
        help=(
            "Threshold for AFT, above which we use a flatted boundary for rise time"
            "Specified values: (AFT boundary, constant rise time)."
        ),
    )

    s1_max_rise_time_post100 = straxen.URLConfig(
        default=200, type=(int, float), help="Maximum S1 rise time for > 100 PE [ns]"
    )

    s1_min_coincidence = straxen.URLConfig(
        default=2, type=int, help="Minimum tight coincidence necessary to make an S1"
    )

    s2_min_pmts = straxen.URLConfig(
        default=4, type=int, help="Minimum number of PMTs contributing to an S2"
    )

    @staticmethod
    def upper_rise_time_area_boundary(area, norm, const, tau):
        """Function which determines the upper boundary for the rise-time for a given area."""
        return norm * np.exp(-area / tau) + const

    @staticmethod
    def upper_rise_time_aft_boundary(aft, slope, offset, aft_boundary, flat_threshold):
        """Function which computes the upper rise time boundary as a function of area fraction
        top."""
        res = 10 ** (slope * aft + offset)
        res[aft >= aft_boundary] = flat_threshold
        return res

    def compute(self, peaklets):
        ptype = np.zeros(len(peaklets), dtype=np.int8)

        # Properties needed for classification:
        rise_time = -peaklets["area_decile_from_midpoint"][:, 1]
        n_channels = (peaklets["area_per_channel"] > 0).sum(axis=1)

        is_large_s1 = peaklets["area"] >= 100
        is_large_s1 &= rise_time <= self.s1_max_rise_time_post100
        is_large_s1 &= peaklets["tight_coincidence"] >= self.s1_min_coincidence

        is_small_s1 = peaklets["area"] < 100
        is_small_s1 &= rise_time < self.upper_rise_time_area_boundary(
            peaklets["area"],
            *self.s1_risetime_area_parameters,
        )

        is_small_s1 &= rise_time < self.upper_rise_time_aft_boundary(
            peaklets["area_fraction_top"],
            *self.s1_risetime_aft_parameters,
            *self.s1_flatten_threshold_aft,
        )

        is_small_s1 &= peaklets["tight_coincidence"] >= self.s1_min_coincidence

        ptype[is_large_s1 | is_small_s1] = 1

        is_s2 = n_channels >= self.s2_min_pmts
        is_s2[is_large_s1 | is_small_s1] = False
        ptype[is_s2] = 2

        peaklets_classification = np.zeros(len(peaklets), dtype=self.dtype)
        peaklets_classification["type"] = ptype
        peaklets_classification["time"] = peaklets["time"]
        peaklets_classification["dt"] = peaklets["dt"]
        peaklets_classification["length"] = peaklets["length"]
        peaklets_classification["channel"] = -1

        return peaklets_classification



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EventWaveform(strax.Plugin):
    """Simple plugin that provides (data) waveforms for main and alternative S1/S2 in the event."""

    depends_on = ("event_basics", "peaks")
    provides = "event_waveform"
    __version__ = "0.0.1"

    compressor = "zstd"
    save_when = strax.SaveWhen.EXPLICIT

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, type=int, help="Number of top PMTs")

    def _have_data(self, field):
        return field in self.deps["peaks"].dtype_for("peaks").names

    def infer_dtype(self):
        # setting data type from peak dtype
        pfields_ = self.deps["peaks"].dtype_for("peaks").fields
        # populating data type
        infoline = {
            "s1": "main S1",
            "s2": "main S2",
            "alt_s1": "alternative S1",
            "alt_s2": "alternative S2",
        }
        dtype = []
        # populating waveform samples
        ptypes = ["s1", "s2", "alt_s1", "alt_s2"]
        for type_ in ptypes:
            dtype += [
                (
                    (f"Waveform for {infoline[type_]} [ PE / sample ]", f"{type_}_data"),
                    pfields_["data"][0],
                )
            ]
            if self._have_data("data_top"):
                dtype += [
                    (
                        (
                            f"Top waveform for {infoline[type_]} [ PE / sample ]",
                            f"{type_}_data_top",
                        ),
                        pfields_["data_top"][0],
                    )
                ]
            if self._have_data("data_start"):
                dtype += [
                    (
                        (
                            f"Waveform Start for {infoline[type_]} [ PE / sample ]",
                            f"{type_}_data_start",
                        ),
                        pfields_["data_start"][0],
                    )
                ]
            dtype += [
                (
                    (f"Length of the interval in samples for {infoline[type_]}", f"{type_}_length"),
                    pfields_["length"][0],
                )
            ]
            dtype += [
                (
                    (f"Width of one sample for {infoline[type_]} [ns]", f"{type_}_dt"),
                    pfields_["dt"][0],
                )
            ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        result = np.zeros(len(events), self.dtype)
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)

        split_peaks = strax.split_by_containment(peaks, events)
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for type_ in ["s1", "s2", "alt_s1", "alt_s2"]:
                type_index = event[f"{type_}_index"]
                if type_index != -1:
                    result[f"{type_}_length"][event_i] = sp["length"][type_index]
                    result[f"{type_}_data"][event_i] = sp["data"][type_index]
                    if self._have_data("data_top"):
                        result[f"{type_}_data_top"][event_i] = sp["data_top"][type_index]
                    if self._have_data("data_start"):
                        result[f"{type_}_data_start"][event_i] = sp["data_start"][type_index]
                    result[f"{type_}_dt"][event_i] = sp["dt"][type_index]
        return result



import numpy as np
import strax
import straxen
from straxen.plugins.hitlets_nv.hitlets_nv import nVETOHitlets
from straxen.plugins.defaults import MV_PREAMBLE, MV_HIT_DEFAULTS

export, __all__ = strax.exporter()


@export
class muVETOHitlets(nVETOHitlets):
    __doc__ = MV_PREAMBLE + (nVETOHitlets.__doc__ or "")
    __version__ = "0.0.2"
    depends_on = "records_mv"

    provides = "hitlets_mv"
    data_kind = "hitlets_mv"
    child_plugin = True

    dtype = strax.hitlet_dtype()

    save_outside_hits_mv = straxen.URLConfig(
        default=MV_HIT_DEFAULTS["save_outside_hits_mv"],
        track=True,
        infer_type=False,
        child_option=True,
        parent_option_name="save_outside_hits_nv",
        help="Save (left, right) samples besides hits; cut the rest",
    )

    hit_min_amplitude_mv = straxen.URLConfig(
        infer_type=False,
        default=MV_HIT_DEFAULTS["hit_min_amplitude_mv"],
        track=True,
        help=(
            "Minimum hit amplitude in ADC counts above baseline. "
            "Specify as a tuple of length n_mveto_pmts, or a number."
        ),
    )

    min_split_mv = straxen.URLConfig(
        default=100,
        track=True,
        infer_type=False,
        child_option=True,
        parent_option_name="min_split_nv",
        help=(
            "Minimum height difference pe/sample between local minimum and maximum, "
            "that a pulse get split."
        ),
    )

    min_split_ratio_mv = straxen.URLConfig(
        default=0,
        track=True,
        infer_type=False,
        child_option=True,
        parent_option_name="min_split_ratio_nv",
        help=(
            "Min ratio between local maximum and minimum to split pulse (zero to switch this off)."
        ),
    )

    gain_model_mv = straxen.URLConfig(
        default=(
            "list-to-array://"
            "xedocs://pmt_area_to_pes"
            "?as_list=True"
            "&sort=pmt"
            "&detector=muon_veto"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
            "&attr=value"
        ),
        infer_type=False,
        child_option=True,
        parent_option_name="gain_model_nv",
        help="PMT gain model. Specify as (model_type, model_config)",
    )

    def setup(self):
        self.channel_range = self.channel_map["mv"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1

        to_pe = self.gain_model_mv
        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0] :] = to_pe[:]

        self.hit_thresholds = self.hit_min_amplitude_mv

    def compute(self, records_mv, start, end):
        return super().compute(records_mv, start, end)



import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class OnlinePeakMonitor(strax.Plugin):
    """Plugin to write data to the online-monitor. Data that is written by this plugin should be
    small such as to not overload the runs- database.

    This plugin takes 'peak_basics' and 'lone_hits'. Although they are not strictly related, they
    are aggregated into a single data_type in order to minimize the number of documents in the
    online monitor.

    Produces 'online_peak_monitor' with info on the lone-hits and peaks

    """

    __version__ = "0.0.5"

    depends_on = ("peak_basics", "lone_hits")
    provides = "online_peak_monitor"
    data_kind = "online_peak_monitor"

    area_vs_width_nbins = straxen.URLConfig(
        type=int,
        default=60,
        help=(
            "Number of bins for area vs width histogram for online monitor. "
            "NB: this is a 2D histogram"
        ),
    )

    area_vs_width_bounds = straxen.URLConfig(
        type=tuple,
        default=((0, 5), (0, 5)),
        help="Boundaries of log-log histogram of area vs width",
    )

    area_vs_width_cut_string = straxen.URLConfig(
        type=str,
        default="",
        help=(
            "Selection (like selection) applied to data for "
            '"area_vs_width_hist_clean", cuts should be separated using "&"'
            "For example: (tight_coincidence > 2) & (area_fraction_top < 0.1)"
            'Default is no selection (other than "area_vs_width_min_gap")'
        ),
    )

    lone_hits_area_bounds = straxen.URLConfig(
        type=tuple, default=(0, 1500), help="Boundaries area histogram of lone hits [ADC]"
    )

    online_peak_monitor_nbins = straxen.URLConfig(
        type=int,
        default=100,
        help=(
            "Number of bins of histogram of online monitor. Will be used "
            "for: "
            "lone_hits_area-histogram, "
            "area_fraction_top-histogram, "
            "online_se_gain estimate (histogram is not stored), "
        ),
    )

    lone_hits_cut_string = straxen.URLConfig(
        type=str,
        default="(area >= 50) & (area <= 250)",
        help=(
            "Selection (like selection) applied to data for "
            '"lone-hits", cuts should be separated using "&")'
        ),
    )

    lone_hits_min_gap = straxen.URLConfig(
        type=int,
        default=15_000,
        help="Minimal gap [ns] between consecutive lone-hits. To turn off this cut, set to 0.",
    )

    n_tpc_pmts = straxen.URLConfig(type=int, help="Number of TPC PMTs")

    online_se_bounds = straxen.URLConfig(
        type=tuple,
        default=(7, 70),
        help="Window for online monitor [PE] to look for the SE gain, value",
    )

    def infer_dtype(self):
        n_bins_area_width = self.area_vs_width_nbins
        bounds_area_width = self.area_vs_width_bounds

        n_bins = self.online_peak_monitor_nbins

        n_tpc_pmts = self.n_tpc_pmts
        dtype = [
            (("Start time of the chunk", "time"), np.int64),
            (("End time of the chunk", "endtime"), np.int64),
            (
                ("Area vs width histogram (log-log)", "area_vs_width_hist"),
                (np.int64, (n_bins_area_width, n_bins_area_width)),
            ),
            (
                ("Area vs width edges (log-space)", "area_vs_width_bounds"),
                (np.float64, np.shape(bounds_area_width)),
            ),
            (("Lone hits areas histogram [ADC-counts]", "lone_hits_area_hist"), (np.int64, n_bins)),
            (("Lone hits areas bounds [ADC-counts]", "lone_hits_area_bounds"), (np.float64, 2)),
            (("Lone hits per channel", "lone_hits_per_channel"), (np.int64, n_tpc_pmts)),
            (("AFT histogram", "aft_hist"), (np.int64, n_bins)),
            (("AFT bounds", "aft_bounds"), (np.float64, 2)),
            (
                ("Number of contributing channels histogram", "n_channel_hist"),
                (np.int64, n_tpc_pmts),
            ),
            (("Single electron gain", "online_se_gain"), np.float32),
        ]
        return dtype

    def compute(self, peaks, lone_hits, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res["time"] = start
        res["endtime"] = end
        n_pmt = self.n_tpc_pmts
        n_bins = self.online_peak_monitor_nbins

        # Bounds for histograms
        res["area_vs_width_bounds"] = self.area_vs_width_bounds
        res["lone_hits_area_bounds"] = self.lone_hits_area_bounds

        # -- Peak vs area 2D histogram --
        # Always cut out unphysical peaks
        sel = (peaks["area"] > 0) & (peaks["range_50p_area"] > 0)
        res["area_vs_width_hist"] = self.area_width_hist(peaks[sel])
        del sel

        # -- Lone hit properties --
        # Make a mask with the cuts.
        # Now only take lone hits that are separated in time.
        if len(lone_hits):
            lh_timedelta = lone_hits[1:]["time"] - strax.endtime(lone_hits)[:-1]
            # Hits on the left are far away? (assume first is because of chunk bound)
            mask = np.hstack([True, lh_timedelta > self.lone_hits_min_gap])
            # Hits on the right are far away? (assume last is because of chunk bound)
            mask &= np.hstack([lh_timedelta > self.lone_hits_min_gap, True])
        else:
            mask = []
        masked_lh = strax.apply_selection(lone_hits[mask], selection=self.lone_hits_cut_string)

        # Make histogram of ADC counts
        # NB: LONE HITS AREA ARE IN ADC!
        lone_hit_areas, _ = np.histogram(
            masked_lh["area"], bins=n_bins, range=self.lone_hits_area_bounds
        )

        lone_hit_channel_count, _ = np.histogram(masked_lh["channel"], bins=n_pmt, range=[0, n_pmt])
        # Count number of lone-hits per PMT
        res["lone_hits_area_hist"] = lone_hit_areas
        res["lone_hits_per_channel"] = lone_hit_channel_count
        # Clear mask, don't re-use
        del mask

        # -- AFT histogram --
        aft_b = [0, 1]
        aft_hist, _ = np.histogram(peaks["area_fraction_top"], bins=n_bins, range=aft_b)
        res["aft_hist"] = aft_hist
        res["aft_bounds"] = aft_b

        # Estimate Single Electron (SE) gain
        se_hist, se_bins = np.histogram(peaks["area"], bins=n_bins, range=self.online_se_bounds)
        bin_centers = (se_bins[1:] + se_bins[:-1]) / 2
        res["online_se_gain"] = bin_centers[np.argmax(se_hist)]
        return res

    def area_width_hist(self, data):
        """Make area vs width 2D-hist."""
        hist, _, _ = np.histogram2d(
            np.log10(data["area"]),
            np.log10(data["range_50p_area"]),
            range=self.area_vs_width_bounds,
            bins=self.area_vs_width_nbins,
        )
        return hist.T



from m2r import convert

from .misc import kind_colors

header = """
Release notes
==============

"""


def convert_release_notes(notes, target, pull_url):
    """Convert the release notes to an RST page with links to PRs."""
    with open(notes, "r") as f:
        notes = f.read()
    rst = convert(notes)
    with_ref = ""
    for line in rst.split("\n"):
        # Get URL for PR
        if "#" in line:
            pr_number = line.split("#")[1]
            while len(pr_number):
                try:
                    pr_number = int(pr_number)
                    break
                except ValueError:
                    # Too many tailing characters to be an int
                    pr_number = pr_number[:-1]
            if pr_number:
                line = line.replace(
                    f"#{pr_number}",
                    f"`#{pr_number} <{pull_url}/{pr_number}>`_",
                )
        with_ref += line + "\n"

    with open(target, "w") as f:
        f.write(header + with_ref)


def add_spaces(x):
    """Add four spaces to every line in x.

    This is needed to make html raw blocks in rst format correctly

    """
    y = ""
    if isinstance(x, str):
        x = x.split("\n")
    for q in x:
        y += "    " + q
    return y


def add_deps_to_graph_tree(graph_tree, plugin, data_type, _seen=None):
    """Recursively add nodes to graph base on plugin.deps."""
    if _seen is None:
        _seen = []
    if data_type in _seen:
        return graph_tree, _seen

    # Add new one
    graph_tree.node(
        data_type,
        style="filled",
        href="#" + data_type.replace("_", "-"),
        fillcolor=kind_colors.get(plugin.data_kind_for(data_type), "grey"),
    )
    for dep in plugin.depends_on:
        graph_tree.edge(data_type, dep)

    # Add any of the lower plugins if we have to
    for lower_data_type, lower_plugin in plugin.deps.items():
        graph_tree, _seen = add_deps_to_graph_tree(graph_tree, lower_plugin, lower_data_type, _seen)
    _seen.append(data_type)
    return graph_tree, _seen



import strax
from straxen.plugins.peaklets.peaklet_positions_cnf import PeakletPositionsCNF

export, __all__ = strax.exporter()


@export
class MergedS2PositionsCNF(PeakletPositionsCNF):

    __version__ = "0.0.0"
    child_plugin = True
    algorithm = "cnf"
    depends_on = "merged_s2s"
    provides = "merged_s2_positions_cnf"

    def compute(self, merged_s2s):
        return super().compute(merged_s2s)



import numpy as np
from immutabledict import immutabledict

import strax
import straxen

from straxen.plugins.defaults import NV_HIT_DEFAULTS

export, __all__ = strax.exporter()


@export
class nVETOHitlets(strax.Plugin):
    """Plugin which computes the nveto hitlets and their parameters.

    Hitlets are an extension of regular hits. They include the left
    and right extension. The plugin does the following:
        1. Generate hitlets which includes these sub-steps:
            * Apply left and right hit extension and concatenate
            overlapping hits.
            * Generate temp. hitelts and look for their waveforms in
            their corresponding records.
            * Split hitlets if they satisfy the set criteria.
        2. Compute the properties of the hitlets.

    Note:
        Hitlets are getting chopped if extended in not recorded regions.

    """

    __version__ = "0.1.1"

    parallel = "process"
    rechunk_on_save = True
    # To reduce the number of chunks, we increase the target size
    # This would not harm memory usage, because we rechunk on load
    chunk_target_size_mb = 2000
    compressor = "zstd"

    depends_on = "records_nv"

    provides = "hitlets_nv"
    data_kind = "hitlets_nv"

    dtype = strax.hitlet_dtype()

    save_outside_hits_nv = straxen.URLConfig(
        default=NV_HIT_DEFAULTS["save_outside_hits_nv"],
        track=True,
        infer_type=False,
        help="Save (left, right) samples besides hits; cut the rest",
    )

    hit_min_amplitude_nv = straxen.URLConfig(
        infer_type=False,
        default=NV_HIT_DEFAULTS["hit_min_amplitude_nv"],
        track=True,
        help=(
            "Minimum hit amplitude in ADC counts above baseline. "
            "Specify as a tuple of length n_nveto_pmts, or a number."
        ),
    )

    min_split_nv = straxen.URLConfig(
        default=0.063,
        track=True,
        infer_type=False,
        help=(
            "Minimum height difference pe/sample between local minimum and maximum, "
            "that a pulse get split."
        ),
    )

    min_split_ratio_nv = straxen.URLConfig(
        default=0.75,
        track=True,
        infer_type=False,
        help=(
            "Min ratio between local maximum and minimum to split pulse (zero to switch this off)."
        ),
    )

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) channel number.",
    )

    gain_model_nv = straxen.URLConfig(
        default=(
            "list-to-array://"
            "xedocs://pmt_area_to_pes"
            "?as_list=True"
            "&sort=pmt"
            "&detector=neutron_veto"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
            "&attr=value"
        ),
        infer_type=False,
        help="PMT gain model. Specify as (model_type, model_config, nT = True)",
    )

    def setup(self):
        self.channel_range = self.channel_map["nveto"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1

        to_pe = self.gain_model_nv

        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0] :] = to_pe[:]

        # Assign attribute that might be used in daughter classes:
        self.hit_thresholds = self.hit_min_amplitude_nv

    def compute(self, records_nv, start, end):
        records_nv = remove_switched_off_channels(records_nv, self.to_pe)
        hits = strax.find_hits(records_nv, min_amplitude=self.hit_thresholds)

        temp_hitlets = strax.create_hitlets_from_hits(
            hits, self.save_outside_hits_nv, self.channel_range, chunk_start=start, chunk_end=end
        )
        del hits

        # Get hitlet data and split hitlets:
        temp_hitlets = strax.get_hitlets_data(
            temp_hitlets, records_nv, to_pe=self.to_pe, min_hitlet_sample=600
        )

        temp_hitlets = strax.split_peaks(
            temp_hitlets,
            None,  # Only needed for peak splitting
            records_nv,
            None,  # Only needed for peak splitting
            self.to_pe,
            data_type="hitlets",
            algorithm="local_minimum",
            min_height=self.min_split_nv,
            min_ratio=self.min_split_ratio_nv,
        )

        # Compute other hitlet properties:
        # We have to loop here 3 times over all hitlets...
        strax.hitlet_properties(temp_hitlets)

        # Remove data field:
        hitlets = np.zeros(len(temp_hitlets), dtype=strax.hitlet_dtype())
        strax.copy_to_buffer(temp_hitlets, hitlets, "_copy_hitlets")
        return strax.sort_by_time(hitlets)


def remove_switched_off_channels(records, to_pe):
    """Removes records of channels which gain was set to zero.

    :param records Hits found in records.
    :param to_pe: conversion factor from ADC per sample.
    :return: records

    """
    channel_off = np.argwhere(to_pe == 0).flatten()
    mask_off = np.isin(records["channel"], channel_off)
    records = records[~mask_off]
    return records



from typing import Union, Tuple
from immutabledict import immutabledict

import numpy as np
import holoviews as hv


# noinspection PyArgumentList
class PlotPeakLikeData:
    """Base class to plot peak like data for any type of detector."""

    # Properties to be overwritten by sub-clases if needed.
    _xlabel = "x-label"  # Label for x axis defined in _set_labels
    _ylabel = "y-label"  # Label for y axis defined in _set_labels
    _label = "label"  # Label for plot defined in _set_labels
    _scaler = 1
    keep_amplitude_per_sample = False
    _vdim_labels = immutabledict()
    _never_include_fields: Union[Tuple[str, ...], Tuple[()]] = ()

    def __init__(
        self,
        opts=immutabledict(),
    ):
        """
        :param opts_dict: Dictionary of options which will be applied to
            hv.Area and hv.Curve plots. If you want to change a setting
            which is not common by both plot types please specify them
            in "plot_peak" or "plot_peaks".
        """
        if not isinstance(opts, (dict, immutabledict)):
            raise ValueError('"opts_dict" must be a dictionary!')
        self.opts_dict = opts

    def plot_peak(self, peak):
        """User defined function for plotting a single peak with correct labels etc."""
        raise NotImplementedError()

    def plot_peaks(self, peaks):
        """User defined function for plotting multiple peaks into the same figure."""
        raise NotImplementedError()

    def _set_labels(self, peaks, peak_type, time_prefix):
        """User defined function which should set the xlabel, ylabel and plot label for our peak
        plots.

        Also should define the scaler which scales the time axis in desired unit e.g. µs or ns.

        """
        raise NotImplementedError()

    def _plot_peak(
        self,
        peak,
        label,
        group_label="Peak",
        opts_curve=immutabledict(),
        opts_area=immutabledict(),
        _relative_start_time=0,
    ):
        """Plots area and curve to display a peak.

        :param peak: Peak to be displayed
        :param label: Label to be used in legend.
        :param opts_curve: Additional options to be applied to hv.Curve
        :param opts_area: Additional options to be applied to hv.Area
        :param _relative_start_time: Time to which the peak should be plotted relatively. E.g. event
            start time.
        :return: hv.Area and hv.Curve

        """
        data = self.get_peak_data(
            peak,
            relative_start=_relative_start_time,
        )

        area = hv.Area(
            data,
            kdims=hv.Dimension("xs", label=self._xlabel),
            vdims=[
                hv.Dimension("ys", label=self._ylabel),
            ],
            label=label,
            group=group_label,
        ).opts(alpha=0.3, **opts_area, framewise=True)

        curve = hv.Curve(
            data,
            kdims=hv.Dimension("xs", label=self._xlabel),
            vdims=[
                hv.Dimension("ys", label=self._ylabel),
            ]
            + self._vdims,
            label=label,
            group=group_label,
        ).opts(**opts_curve, framewise=True)

        return area, curve

    def _init_peak_plot(self, peak, peak_type, time_prefix):
        """Initalizes empty plots for plotting single or muliple-peaks into the same figure.

        Intializes plot legend labels as well as other dimensions.

        """
        self._set_single_valued_dimensions(peak)
        self._set_labels(peak, peak_type, time_prefix)

        area = hv.Area(
            None,
            kdims=hv.Dimension("xs", label=self._xlabel),
            vdims=[
                hv.Dimension("ys", label=self._ylabel),
            ],
        ).opts(framewise=True)

        curve = hv.Curve(
            None,
            kdims=hv.Dimension("xs", label=self._xlabel),
            vdims=[
                hv.Dimension("ys", label=self._ylabel),
            ]
            + self._vdims,
        ).opts(framewise=True)
        return area, curve

    def _set_single_valued_dimensions(self, peak):
        """Extracts information from peaks. Defines a list of single valued dimensions which will be
        extracted for the plot in _get_peak_data.

        Defines holovies vdim Dimensions for a proper hover-tool display. Looks up unit describition
        in _vdim_labels which have to be specified by the user.

        """
        self.single_dimension_fields = self._get_single_dimension_fields(peak)
        vdims = []

        for field_name in self.single_dimension_fields:
            if field_name in self._never_include_fields:
                continue
            vdims.append(
                hv.Dimension(field_name, label=self._vdim_labels.get(field_name, field_name))
            )
        self._vdims = vdims

    @staticmethod
    def _get_single_dimension_fields(peak):
        """Function which returns all single valued qunatity names for a given peak."""
        field_names = []
        for field_name in peak.dtype.names:
            value = peak[field_name]
            _is_single_value_quantity = value.ndim == 1
            if _is_single_value_quantity:
                field_names.append(field_name)

        return field_names

    def _is_time_based_quantity(self, field_name):
        """Function which checks if specified parameter is a quantiy which is based on time, like
        range, width or time properties."""
        raise NotImplementedError()

    def get_peak_data(self, peak, relative_start=0):
        """Function which extracts base peaks information as a dictionary. Can be further customized
        in sub-classes if needed.

        :param peak: Peaks to be plotted.
        :param relative_start: t0 from which on the peaks should be plotted.
        :return: dictionary

        """
        # Wrapper in case one hast to customized things further in one
        # of the sub-clases.
        return self._get_peak_data(peak, relative_start=relative_start)

    def _get_peak_data(self, peak, relative_start=0):
        """Function which extracts base peaks information as a dictionary. Extracts all information
        which are stored as single valued parameters.

        :param peaks: Peaks to be plotted.
        :param relative_start: t0 from which on the peaks should be plotted.
        :return: dictionary

        """
        x, y = self._patches_x_y(peak)
        x -= relative_start  # relative to first peak
        x = x / self._scaler  # Scaler to convert ns to µs or other units

        # Coordinates for peak patches
        data = {
            "xs": x,
            "ys": y,
        }

        # Add other single valued fields which we can find to our
        # dictionary. Adjust time parameters according to scaling.
        for field_name in self.single_dimension_fields:
            value = peak[field_name]

            if self._is_time_based_quantity(field_name):
                data[field_name] = value / self._scaler
            else:
                data[field_name] = value

        return data

    def _patches_x_y(self, peak):
        """Creates x,y coordinates needed to make a stepwise function with hv.Areas.

        :param peak: Peak for which we need the x/y samples
        :return: Tuple of x, y

        """
        if self.keep_amplitude_per_sample:
            dt_a = 1
        else:
            dt_a = peak["dt"]

        x = np.arange(peak["length"])
        xx = np.zeros(2 * len(x), dtype=x.dtype)
        mx = 0.5 * (x[1::] + x[:-1])
        xx[1:-1:2] = mx
        xx[2::2] = mx
        xx[0] = 1.5 * x[0] - 0.5 * x[1]
        xx[-1] = 1.5 * x[-1] - 0.5 * x[-2]
        xx = np.array(xx) * peak["dt"] + peak["time"]
        y = peak["data"][: peak["length"]]
        yy = np.zeros(2 * len(x))
        yy[0::2] = y
        yy[1::2] = y
        yy = np.array(yy) / dt_a

        # baseline since we'll fill underneath
        xx = np.concatenate([[xx[0]], xx, [xx[-1]]])
        yy = np.concatenate([[0], yy, [0]])
        return xx, yy


# noinspection PyArgumentList
class PlotPeaksTPC(PlotPeakLikeData):
    """Class which plots single or multiple peaks for interactive holoviews display."""

    time_in_us = False
    _never_include_fields = (
        "time",
        "endtime",
        "center_time",
        "channel",
        "type",
        "max_goodness_of_split",
    )
    # Use round brakets for units which should not be affected by
    # "time_in_µs".
    _vdim_labels = {
        "time": "time (ns)",
        "endtime": "endtime (ns)",
        "center_time": "center_time (ns)",
        "rise_time": "rise_time [ns]",
        "range_50p_area": "range_50p_area [ns]",
        "range_90p_area": "range_90p_area [ns]",
        "area": "area [PE]",
        "max_pmt_area": "max_pmt_area [PE]",
    }

    def __init__(
        self,
        opts=immutabledict(),
        keep_amplitude_per_sample=False,
    ):
        super().__init__(opts)
        self.keep_amplitude_per_sample = keep_amplitude_per_sample

    def _set_labels(self, peak, peak_type, time_prefix):
        if self.time_in_us:
            scaler = 1000
            xlabel = f"{time_prefix} Time [µs]"
            self._change_units()
        else:
            xlabel = f"{time_prefix} Time [ns]"
            scaler = 1

        if self.keep_amplitude_per_sample:
            ylabel = f"{peak_type} Amplitude [pe/sample]"
        else:
            ylabel = f"{peak_type} Amplitude [pe/ns]"

        self._scaler = scaler
        self._xlabel = xlabel
        self._ylabel = ylabel
        self._label = f'S{peak["type"]}'

    def _change_units(self):
        for dim in self._vdims:
            dim.label = dim.label.replace("[ns]", "[µs]")

    def plot_peak(
        self,
        peak,
        opts_area=immutabledict(),
        opts_curve=immutabledict(),
        label="Peak",
        time_in_us=False,
        amplitude_prefix="",
    ):
        """Function which plots a single peak."""
        plot = self.plot_peaks(
            peak,
            opts_curve=opts_curve,
            opts_area=opts_area,
            time_in_us=time_in_us,
            label=label,
            group_label="Peak",
            amplitude_prefix=amplitude_prefix,
            _relative_start_time=peak["time"],
        )
        # FIY: If opts is applied only with an empty dict all previous
        # options seem to be reseted to default.
        return plot.opts(**self.opts_dict, legend_limit=10)

    def plot_peaks(
        self,
        peaks,
        label="Peaks",
        group_label="Event",
        opts_area=immutabledict(),
        opts_curve=immutabledict(),
        time_in_us=False,
        time_prefix="",
        amplitude_prefix="",
        _relative_start_time=0,
    ):
        if type(peaks) is np.void:
            # In case of supplying only a single peak we have to make
            # an array again:
            peaks = np.array([peaks])

        if not (np.all(peaks["type"] == peaks[0]["type"])):
            raise ValueError("All peaks must be of the same type (S1, S2 or Unknown)!")

        self.time_in_us = time_in_us
        area, curve = self._init_peak_plot(peaks, amplitude_prefix, time_prefix)
        for ind, peak in enumerate(peaks):
            _area, _curve = self._plot_peak(
                peak,
                label=label + f"_{ind}",
                group_label=group_label,
                opts_curve=opts_curve,
                opts_area=opts_area,
                _relative_start_time=_relative_start_time,
            )
            area *= _area
            curve *= _curve.opts(tools=["vline"])
        return (area * curve).opts(**self.opts_dict, legend_limit=100)

    def _is_time_based_quantity(self, field_name):
        """Checks if fiel_name belongs to a time based peak property to rescale ns value to µs for
        S2s."""
        _is_true = "time" in field_name and (
            (field_name != "time") and (field_name != "endtime") and (field_name != "center_time")
        )
        _is_true |= "width" in field_name
        _is_true |= "range" in field_name
        return _is_true



import matplotlib.pyplot as plt
import numpy as np
import straxen
from multihist import Hist1d, Histdd
from matplotlib.colors import LogNorm


@straxen.mini_analysis(requires=("peak_basics",))
def plot_peaks_aft_histogram(
    context,
    run_id,
    peaks,
    pe_bins=np.logspace(0, 7, 120),
    rt_bins=np.geomspace(2, 1e5, 120),
    extra_labels=tuple(),
    rate_range=(1e-4, 1),
    aft_range=(0, 0.85),
    figsize=(14, 5),
):
    """Plot side-by-side (area, width) histograms of the peak rate and mean area fraction top.

    :param pe_bins: Array of bin edges for the peak area dimension [PE]
    :param rt_bins: array of bin edges for the rise time dimension [ns]
    :param extra_labels: List of (area, risetime, text, color) extra labels to put on the plot
    :param rate_range: Range of rates to show [peaks/(bin*s)]
    :param aft_range: Range of mean S1 area fraction top / bin to show
    :param figsize: Figure size to use

    """
    livetime_sec = straxen.get_livetime_sec(context, run_id, peaks)

    mh = Histdd(
        peaks,
        dimensions=(
            ("area", pe_bins),
            ("range_50p_area", rt_bins),
            ("area_fraction_top", np.linspace(0, 1, 100)),
        ),
    )

    f, axes = plt.subplots(1, 2, figsize=figsize)

    def std_axes():
        plt.gca().set_facecolor("k")
        plt.yscale("log")
        plt.xscale("log")
        plt.xlabel("Area [PE]")
        plt.ylabel("Range 50% area [ns]")
        labels = [
            (12, 8, "AP?", "white"),
            (3, 150, "1PE\npileup", "gray"),
            (30, 200, "1e", "gray"),
            (100, 1000, "n-e", "w"),
            (2000, 2e4, "Train", "gray"),
            (1200, 50, "S1", "w"),
            (45e3, 60, "αS1", "w"),
            (2e5, 800, "S2", "w"),
        ] + list(extra_labels)

        for x, w, text, color in labels:
            plt.text(
                x, w, text, color=color, verticalalignment="center", horizontalalignment="center"
            )

    plt.sca(axes[0])
    (mh / livetime_sec).sum(axis=2).plot(
        norm=LogNorm(vmin=rate_range[0], vmax=rate_range[1]),
        colorbar_kwargs=dict(extend="both"),
        cblabel="Peaks / (bin * s)",
    )
    std_axes()

    plt.sca(axes[1])
    mh.average(axis=2).plot(
        vmin=aft_range[0],
        vmax=aft_range[1],
        colorbar_kwargs=dict(extend="max"),
        cmap=plt.cm.jet,
        cblabel="Mean area fraction top",
    )

    std_axes()
    plt.tight_layout()


@straxen.mini_analysis(requires=["event_info"])
def event_scatter(
    context,
    run_id,
    events,
    show_single=True,
    s=10,
    color_range=(None, None),
    color_dim="s1_area_fraction_top",
    figsize=(7, 5),
):
    """Plot a (cS1, cS2) event scatter plot.

    :param show_single: Show events with only S1s or only S2s just besides the axes.
    :param s: Scatter size
    :param color_dim: Dimension to use for the color. Must be in event_info.
    :param color_range: Minimum and maximum color value to show.
    :param figsize: (w, h) figure size to use, or leave None to not make a new matplotlib figure.

    """
    if figsize is not None:
        plt.figure(figsize=figsize)
    if color_dim == "s1_area_fraction_top" and color_range == (None, None):
        color_range = (0, 0.3)

    plt.scatter(
        np.nan_to_num(events["cs1"]).clip(0.9, None),
        np.nan_to_num(events["cs2"]).clip(0.9, None),
        clip_on=not show_single,
        c=events[color_dim],
        vmin=color_range[0],
        vmax=color_range[1],
        s=s,
        cmap=plt.cm.jet,
        marker=".",
        edgecolors="none",
    )

    plt.xlabel("cS1 [PE]")
    plt.xscale("log")
    plt.xlim(1, None)

    plt.ylabel("cS2 [PE]")
    plt.yscale("log")
    plt.ylim(1, None)

    p = context.get_single_plugin(run_id, "energy_estimates")
    ax = plt.gca()
    el_lim = p.cs1_to_e(np.asarray(ax.get_xlim()))
    ec_lim = p.cs2_to_e(np.asarray(ax.get_ylim()))

    ax2 = ax.twiny()
    ax2.set_xlim(*el_lim)
    ax2.set_xscale("log")
    ax2.set_xlabel("E_light [keVee]")

    ax3 = ax2.twinx()
    ax3.set_ylim(*ec_lim)
    ax3.set_yscale("log")
    ax3.set_ylabel("E_charge [keVee]")

    plt.sca(ax3)
    plt.plot(el_lim, el_lim, c="k", alpha=0.2)
    x = np.geomspace(*el_lim, num=1000)
    e_label = 1.2e-3
    for e_const, label in [
        (0.1, ""),
        (1, "1\nkeV"),
        (10, "10\nkeV"),
        (100, "100\nkeV"),
        (1e3, "1\nMeV"),
        (1e4, ""),
    ]:
        plt.plot(x, e_const - x, c="k", alpha=0.2)
        plt.text(
            e_const - e_label,
            e_label,
            label,
            bbox=dict(facecolor="white", alpha=0.5, edgecolor="none"),
            horizontalalignment="center",
            verticalalignment="center",
            color="k",
            alpha=0.5,
        )

    plt.sca(ax)
    if color_range[0] is None:
        extend = "neither" if color_range[1] is None else "max"
    else:
        extend = "min" if color_range[1] is None else "both"
    if color_dim == "s1_area_fraction_top":
        plt.colorbar(label="S1 area fraction top", extend=extend, ax=[ax, ax3])
    else:
        plt.colorbar(label=color_dim, extend=extend, ax=[ax, ax3])


@straxen.mini_analysis(requires=("event_info",))
def plot_energy_spectrum(
    events,
    color="b",
    label=None,
    unit=None,
    exposure_kg_sec=None,
    error_alpha=0.5,
    errors="fc",
    n_bins=100,
    min_energy=1,
    max_energy=100,
    geomspace=True,
):
    """Plot an energy spectrum histogram, with 1 sigma Poisson confidence intervals around it.

    :param exposure_kg_sec: Exposure in kg * sec
    :param unit: Unit to plot spectrum in. Can be either:
        - events (events per bin)
        - kg_day_kev (events per kg day keV)
        - tonne_day_kev (events per tonne day keV)
        - tonne_year_kev (events per tonne year keV)
        Defaults to kg_day_kev if exposure_kg_sec is provided, otherwise events.
    :param min_energy: Minimum energy of the histogram
    :param max_energy: Maximum energy of the histogram
    :param geomspace: If True, will use a logarithmic energy binning.
        Otherwise will use a linear scale.
    :param n_bins: Number of energy bins to use
    :param color: Color to plot in
    :param label: Label for the line
    :param error_alpha: Alpha value for the statistical error band
    :param errors: Type of errors to draw, passed to 'errors'
        argument of Hist1d.plot.

    """
    if unit is None:
        if exposure_kg_sec is not None:
            unit = "kg_day_kev"
        else:
            unit = "events"

    h = Hist1d(
        events["e_ces"],
        bins=(np.geomspace if geomspace else np.linspace)(min_energy, max_energy, n_bins),
    )

    if unit == "events":
        scale, ylabel = 1, "Events per bin"
    else:
        if exposure_kg_sec is None:
            raise ValueError("you did not specify exposure_kg_sec")
        exposure_kg_day = exposure_kg_sec / (3600 * 24)
        if unit == "kg_day_kev":
            scale = exposure_kg_day
            ylabel = "Events / (kg day keV)"
        elif unit == "tonne_day_kev":
            scale = exposure_kg_day / 1000
            ylabel = "Events / (tonne day keV)"
        elif unit == "tonne_year_kev":
            scale = exposure_kg_day / 1000
            ylabel = "Events / (tonne year keV)"
        else:
            raise ValueError(f"Invalid unit {unit}")
        scale *= h.bin_volumes()

    h.plot(
        errors=errors,
        error_style="band",
        color=color,
        label=label,
        linewidth=1,
        scale_histogram_by=1 / scale,
        error_alpha=error_alpha,
    )
    plt.yscale("log")
    if geomspace:
        straxen.log_x(min_energy, max_energy, scalar_ticks=True)
    else:
        plt.xlim(min_energy, max_energy)
    plt.ylabel(ylabel)
    plt.xlabel("Energy [keV_ee], CES")


@straxen.mini_analysis(requires=("peak_basics",))
def plot_peak_classification(peaks, s=1):
    """Make an (area, rise_time) scatter plot of peaks.

    :param s: Size of dot for each peak

    """
    for cl, color in enumerate("kbg"):
        d = peaks[peaks["type"] == cl]
        plt.scatter(
            d["area"],
            d["rise_time"],
            c=color,
            s=s,
            marker=".",
            edgecolors="none",
            label={0: "Unknown", 1: "S1", 2: "S2"}[cl],
        )
    plt.legend(loc="lower right", markerscale=10)

    plt.xscale("log")
    plt.yscale("log")
    plt.xlim(1, 2e6)
    plt.ylim(3, 1e4)
    plt.xlabel("Area [PE]")
    plt.ylabel("Rise time [ns]")



import os
import re
import typing
import socket
from typing import Dict, Union, Iterator
from copy import deepcopy
import strax
from .rucio_remote import key_to_rucio_did
from .rucio_local import RucioLocalBackend

import warnings

try:
    import utilix
except (RuntimeError, FileNotFoundError):
    # We might be on a travis job
    pass
from straxen import uconfig

export, __all__ = strax.exporter()


@export
class RunDB(strax.StorageFrontend):
    """Frontend that searches RunDB MongoDB for data."""

    storage_type = strax.StorageType.LOCAL
    # Dict of alias used in rundb: regex on hostname
    hosts = {
        "dali": r"^dali.*rcc.*|fried.rice.edu",
        "midway": r"^midway2.*rcc.*|^midway.*rcc.*",
    }
    userdisks = {
        "UC_DALI_USERDISK": r"^dali.*rcc.*|fried.rice.edu",
        "UC_MIDWAY_USERDISK": r"^midway2.*rcc.*|^midway.*rcc.*",
    }

    provide_run_metadata = True
    progress_bar = False

    def __init__(
        self,
        minimum_run_number=7157,
        maximum_run_number=None,
        runid_field="name",
        local_only=False,
        new_data_path=None,
        reader_ini_name_is_mode=False,
        rucio_path=None,
        mongo_url=None,
        mongo_user=None,
        mongo_password=None,
        mongo_database=None,
        *args,
        **kwargs,
    ):
        """
        :param minimum_run_number: Lowest number to consider
        :param maximum_run_number: Highest number to consider. When None
            (the default) consider all runs that are higher than the
            minimum_run_number.
        :param runid_field: Rundb field to which strax's run_id concept
            corresponds. Can be either
            - 'name': values must be strings, for XENON1T
            - 'number': values must be ints, for XENONnT DAQ tests
        :param local_only: Do not show data as available if it would
            have to be downloaded from a remote location.
        :param new_data_path: Path where new files are to be written.
            Defaults to None: do not write new data
            New files will be registered in the runs db!
        :param reader_ini_name_is_mode: If True, will overwrite the
            'mode' field with 'reader.ini.name'.
        :param rucio_path: What is the base path where Rucio is mounted
        :param mongo_url: URL to Mongo runs database (excl auth)
        :param mongo_user: user to Mongo runs database
        :param mongo_password: password to Mongo runs database
        :param mongo_database: database name of Mongo runs database

        Other (kw)args are passed to StorageFrontend.__init__
        """
        super().__init__(*args, **kwargs)
        self.local_only = local_only
        self.new_data_path = new_data_path
        self.reader_ini_name_is_mode = reader_ini_name_is_mode
        self.minimum_run_number = minimum_run_number
        self.maximum_run_number = maximum_run_number
        self.rucio_path = rucio_path
        if self.new_data_path is None:
            self.readonly = True
        self.runid_field = runid_field

        if self.runid_field not in ["name", "number"]:
            raise ValueError("Unrecognized runid_field option %s" % self.runid_field)

        self.hostname = socket.getfqdn()
        if not self.readonly and self.hostname.endswith("xenon.local"):
            # We want admin access to start writing data!
            mongo_url = uconfig.get("rundb_admin", "mongo_rdb_url")
            mongo_user = uconfig.get("rundb_admin", "mongo_rdb_username")
            mongo_password = uconfig.get("rundb_admin", "mongo_rdb_password")
            mongo_database = uconfig.get("rundb_admin", "mongo_rdb_database")

        # setup mongo kwargs...
        # utilix.rundb.xent_collection will take the following variables as kwargs
        #     url: mongo url, including auth
        #     user: the user
        #     password: the password for the above user
        #     database: the mongo database name
        # finally, it takes the collection name as an arg (not a kwarg).
        # if no collection arg is passed, it defaults to the RunDB collection
        # See github.com/XENONnT/utilix/blob/master/utilix/rundb.py for more details
        mongo_kwargs = {
            "url": mongo_url,
            "user": mongo_user,
            "password": mongo_password,
            "database": mongo_database,
        }
        self.collection = utilix.rundb.xent_collection(**mongo_kwargs)

        # Do not delete the client!
        self.client = self.collection.database.client

        self.backends = [
            strax.FileSytemBackend(),
        ]

        # Construct mongo query for runs with available data.
        # This depends on the machine you're running on.
        self.available_query = [{"host": self.hostname}]

        # Go through known host aliases
        for host_alias, regex in self.hosts.items():
            if re.match(regex, self.hostname):
                self.available_query.append({"host": host_alias})

        # When querying for rucio, add that it should be sparsed by the hostname
        if self.rucio_path is not None and any(
            re.match(regex, self.hostname) for regex in self.hosts.values()
        ):
            for host_alias, regex in self.userdisks.items():
                if re.match(regex, self.hostname):
                    self.backends.append(RucioLocalBackend(self.rucio_path))
                    self.available_query.append(
                        {
                            "host": "rucio-catalogue",
                            "location": host_alias,
                            "status": "transferred",
                        }
                    )

    def _data_query(self, key):
        """Return MongoDB query for data field matching key."""
        return {
            "data": {
                "$elemMatch": {
                    "type": key.data_type,
                    "$and": [
                        {
                            "$or": [
                                {"meta.lineage": key.lineage},
                                {
                                    "did": {"$regex": f"/*{key.data_type}-{key.lineage_hash}"},
                                },
                            ]
                        },
                        {"$or": self.available_query},
                    ],
                }
            }
        }

    def _find(self, key: strax.DataKey, write, allow_incomplete, fuzzy_for, fuzzy_for_options):
        if key.run_id.startswith("_"):
            # Superruns are currently not supprorted..
            raise strax.DataNotAvailable

        if fuzzy_for or fuzzy_for_options:
            warnings.warn("Can't do fuzzy with RunDB yet. Only returning exact matches")

        # Check if the run exists
        run_query: Dict[str, Union[str, int]]
        if self.runid_field == "name":
            run_query = {"name": str(key.run_id)}
        else:
            run_query = {"number": int(key.run_id)}

        # Check that we are in rucio backend
        if self.rucio_path is not None:
            rucio_key = key_to_rucio_did(key)
            rucio_available_query = self.available_query[-1]
            dq = {
                "data": {
                    "$elemMatch": {
                        "type": key.data_type,
                        "did": rucio_key,
                        **rucio_available_query,
                    },
                }
            }
            doc = self.collection.find_one(
                {
                    **run_query,
                    **dq,
                },
                projection=dq,
            )
            if doc is not None:
                datum = doc["data"][0]
                error_message = f'Expected {rucio_key} got data on {datum["location"]}'
                if datum.get("did", "") != rucio_key:
                    raise RuntimeError(error_message)
                backend_name = "RucioLocalBackend"
                backend_key = key_to_rucio_did(key)
                return backend_name, backend_key

        dq = self._data_query(key)
        doc = self.collection.find_one({**run_query, **dq}, projection=dq)

        if doc is None:
            # Data was not found
            if not write:
                raise strax.DataNotAvailable

            output_path = os.path.join(self.new_data_path, str(key))

            if self.new_data_path is not None:
                doc = self.collection.find_one(run_query, projection={"_id"})
                if not doc:
                    raise ValueError(
                        f"Attempt to register new data for non-existing run {key.run_id}"
                    )
                self.collection.find_one_and_update(
                    {"_id": doc["_id"]},
                    {
                        "$push": {
                            "data": {
                                "location": output_path,
                                "host": self.hostname,
                                "type": key.data_type,
                                "protocol": strax.FileSytemBackend.__name__,
                                "meta": {"lineage": key.lineage},
                            }
                        }
                    },
                )

            return (strax.FileSytemBackend.__name__, output_path)
        datum = doc["data"][0]

        if datum["host"] == "rucio-catalogue":
            raise strax.DataNotAvailable

        if write and not self._can_overwrite(key):
            raise strax.DataExistsError(at=datum["location"])

        return datum["protocol"], datum["location"]

    def find_several(self, keys: typing.List[strax.DataKey], **kwargs):
        if kwargs.get("fuzzy_for", False) or kwargs.get("fuzzy_for_options", False):
            warnings.warn("Can't do fuzzy with RunDB yet. Only returning exact matches")
        if not keys:
            return []
        if not len(set([k.lineage_hash for k in keys])) == 1:
            raise ValueError("find_several keys must have same lineage")
        if not len(set([k.data_type for k in keys])) == 1:
            raise ValueError("find_several keys must have same data type")
        keys = list(keys)  # Context used to pass a set

        if self.runid_field == "name":
            run_query = {"name": {"$in": [key.run_id for key in keys]}}
        else:
            run_query = {
                f"{self.runid_field}": {
                    "$in": [int(key.run_id) for key in keys if not key.run_id.startswith("_")]
                }
            }
        dq = self._data_query(keys[0])

        # dict.copy is sometimes not sufficient for nested dictionary
        projection = deepcopy(dq)
        projection.update({k: True for k in f"name number".split()})
        results_dict = self._parse_documents(
            self.collection.find({**run_query, **dq}, projection=projection),
        )

        return [results_dict.get(k.run_id, False) for k in keys]

    def _parse_documents(self, documents: Iterator):
        results_dict = dict()
        for doc in documents:
            # If you get a key error here there might be something off with the
            # projection
            datum = doc["data"][0]

            if self.runid_field == "name":
                dk = doc["name"]
            else:
                dk = f'{doc["number"]:06}'
            for required_field in ("protocol", "location"):
                if required_field not in datum:
                    raise ValueError(f"Missing field {required_field} in {datum} from {doc}")
            results_dict[dk] = datum["protocol"], datum["location"]
        return results_dict

    def _scan_runs(self, store_fields):
        query = self.number_query()
        projection = strax.to_str_tuple(list(store_fields))
        # Replace fields by their subfields if requested only take the most
        # "specific" projection
        projection = [
            f1 for f1 in projection if not any([f2.startswith(f1 + ".") for f2 in projection])
        ]
        cursor = self.collection.find(filter=query, projection=projection)
        for doc in strax.utils.tqdm(
            cursor,
            desc="Fetching run info from MongoDB",
            total=self.collection.count_documents(query),
            disable=not self.progress_bar,
        ):
            del doc["_id"]
            if self.reader_ini_name_is_mode:
                doc["mode"] = doc.get("reader", {}).get("ini", {}).get("name", "")
            yield doc

    def run_metadata(self, run_id, projection=None):
        if run_id.startswith("_"):
            # Superruns are currently not supported..
            raise strax.DataNotAvailable

        if self.runid_field == "name":
            run_id = str(run_id)
        else:
            run_id = int(run_id)
        if isinstance(projection, str):
            projection = {projection: 1}
        elif isinstance(projection, (list, tuple)):
            projection = {x: 1 for x in projection}

        doc = self.collection.find_one({self.runid_field: run_id}, projection=projection)
        if doc is None:
            raise strax.DataNotAvailable
        if self.reader_ini_name_is_mode:
            doc["mode"] = doc.get("reader", {}).get("ini", {}).get("name", "")
        return doc

    def number_query(self):
        q_number = {}
        if self.minimum_run_number:
            q_number = {"$gt": self.minimum_run_number}
        if self.maximum_run_number is not None:
            q_number.update({"$lt": self.maximum_run_number})
        if q_number:
            return {"number": q_number}
        return {}



import strax
from straxen.plugins.peaks._peak_positions_base import PeakPositionsBase

export, __all__ = strax.exporter()


@export
class PeakletPositions(PeakPositionsBase):

    __version__ = "0.0.0"
    child_plugin = True
    provides = "peaklet_positions"
    depends_on = (
        "peaklet_positions_cnf",
        "peaklet_positions_mlp",
    )

    def compute(self, peaklets):
        return super().compute(peaklets)



import strax
import numpy as np

from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.merged_s2s.merged_s2s import MergedS2s

export, __all__ = strax.exporter()


@export
class MergedS2sHighEnergy(MergedS2s):
    __doc__ = HE_PREAMBLE + (MergedS2s.__doc__ or "")
    depends_on = ("peaklets_he", "peaklet_classification_he")
    data_kind = "merged_s2s_he"
    provides = "merged_s2s_he"
    __version__ = "0.1.0"
    child_plugin = True

    def infer_dtype(self):
        return strax.unpack_dtype(self.deps["peaklets_he"].dtype_for("peaklets_he"))

    @property
    def use_bayesian_merging(self):
        # there is no pos-rec for HE channels
        return False

    def compute(self, peaklets_he, start, end):
        if self.no_merging(peaklets_he):
            return self.empty_result()

        _peaklets_he = strax.merge_arrs(
            [peaklets_he, np.zeros(len(peaklets_he), dtype=self.indicator_dtype)],
            dtype=strax.merged_dtype((peaklets_he.dtype, self.indicator_dtype)),
        )

        # There are not any lone hits for the high energy channel,
        # so create a dummy for the compute method.
        lone_hits = np.zeros(0, dtype=strax.hit_dtype)
        return self.merge(_peaklets_he, lone_hits, start, end)[0]
